import numpy as np
from astropy.table import Table
import os
import matplotlib.pyplot as plt
import pandas as pd
from astropy.io import fits
import csv
from astropy.io import ascii
from astropy.table import Table
import requests
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.coordinates import Angle
#from astro_ghost import PS1QueryFunctions as ps1
from astropy.utils.data import get_pkg_data_filename
from matplotlib import colors
from imblearn.under_sampling import RandomUnderSampler
from scipy import ndimage
from astropy.wcs import WCS
from matplotlib.pyplot import figure
import pickle
import re
from sklearn import preprocessing
from sklearn.model_selection import train_test_split
from sklearn.metrics import roc_curve, auc
from sklearn.model_selection import StratifiedKFold
from scipy import interp
import seaborn as sns
from collections import Counter
from imblearn.datasets import make_imbalance
import random
from imblearn.over_sampling import SMOTE
from sklearn import preprocessing
from imblearn.pipeline import Pipeline
from rfpimp import *

def plot_ROC_wCV_wMandel(foleySN_matrix_imputed, foleylabels, dataML_matrix_imputed, labels, save=1, balance=1):
    sns.set_context('paper')
    fig = plt.figure(figsize=(10,10), frameon=False)
    ax = plt.gca()
    accuracyF = plot_ROC_wCV(ax, foleySN_matrix_imputed, foleylabels, save=1, balance=balance)
    accuracyM = plot_ROC_wCV(ax, dataML_matrix_imputed, labels, save=0, balance=balance)
    plt.xlabel("False Positive Rate", fontsize=16);
    plt.ylabel("True Positive Rate", fontsize=16);
    plt.legend(loc=4, fontsize=20)

    if save:
        plt.savefig("Combined_MeanROC_Curve_%i_Classes_dataML.png" % len(classes))
    else:
        plt.show()

def plot_ROC_wCV(ax, X, y, names, save=True, balance=True):
    sns.set_context("paper")
    nsplit = 5
    cv = StratifiedKFold(n_splits=nsplit)
    classes = np.unique(y)
    colors = plt.cm.Dark2(np.linspace(0, 1, len(classes)))
    rf = RandomForestClassifier(n_estimators=1400, min_samples_split=2, min_samples_leaf=1, max_features='sqrt', max_depth=90, bootstrap=False)
    tprs = []
    allAcc = []
    aucs = []
    all_confMatrices = []
    mean_fpr = np.linspace(0, 1, 100)
    accuracy_tot = 0
    nclass = len(classes)
    wrong = []
    for j in range(nclass):
        i = 0
        for train, test in cv.split(X, y):
            names_test = names[test]
            if nclass == 3:
                sampling1={'SN Ia': Counter(y[train])['SN Ia'], 'Core Collapse': Counter(y[train])['Core Collapse'], 'SLSN': 1000}
                sampling2={'SN Ia': 1000, 'SLSN': 1000, 'Core Collapse': 1000}
            elif nclass == 4:
                sampling1={'SN Ia':Counter(y[train])['SN Ia'], 'Core Collapse':Counter(y[train])['Core Collapse'] , 'SN Ia Pec': 500, 'SLSN': 500}
                sampling2={'SN Ia': 1000, 'Core Collapse': 1000, 'SN Ia Pec': 500, 'SLSN': 500}
            elif nclass == 2:
                sampling1={'SN Ia': Counter(y[train])['SN Ia'], 'Core Collapse': 3500}
                sampling2={'SN Ia': 3500, 'Core Collapse': 3500}
            if balance:
                over = SMOTE(sampling_strategy=sampling1)
                under = RandomUnderSampler(sampling_strategy=sampling2)
                steps = [('o', over), ('u', under)]
                pipeline = Pipeline(steps=steps)
                Xtrain_resampled, ytrain_resampled = pipeline.fit_resample(X[train], y[train])
            else:
                Xtrain_resampled = X[train]
                ytrain_resampled = y[train]
            print('Distribution after imbalancing: {}'.format(Counter(ytrain_resampled)))
            print('Distribution of test set: {}'.format(Counter(y[test])))

            probas_ = rf.fit(Xtrain_resampled, ytrain_resampled).predict_proba(X[test])
            predictions = rf.predict(X[test])
            # Compute ROC curve and area the curve
            fpr, tpr, thresholds = roc_curve(y[test], probas_[:, j], pos_label=classes[j])
            tprs.append(interp(mean_fpr, fpr, tpr))
            tprs[-1][0] = 0.0
            roc_auc = auc(fpr, tpr)
            aucs.append(roc_auc)
            i += 1
            tempAccuracy =  np.sum(predictions == y[test])/len(y[test])*100
            wrong.append(names_test[y[test] != predictions])
            print(tempAccuracy)
            allAcc.append(tempAccuracy)
            matr = sklearn.metrics.confusion_matrix(y[test], predictions, normalize='true')
            all_confMatrices.append(matr)
            print(matr)
            accuracy_tot += tempAccuracy
        mean_tpr = np.mean(tprs, axis=0)
        mean_tpr[-1] = 1.0
        mean_auc = auc(mean_fpr, mean_tpr)
        std_auc = np.std(aucs)
        accuracy = accuracy_tot / (nsplit*len(classes))
        if True:
            if classes[j] == 'Core Collapse':
                ax.plot(mean_fpr, mean_tpr, color=colors[j],
                         label=r'CC (%0.2f $\pm$ %0.2f)' % (mean_auc, std_auc),
                         lw=2, alpha=.8)
            elif classes[j] == 'SLSN':
                ax.plot(mean_fpr, mean_tpr, color=colors[j],
                         label=r'%s (%0.2f $\pm$ %0.2f)' % (classes[j], mean_auc, std_auc),
                         lw=2, alpha=.8)
            else:
                ax.plot(mean_fpr, mean_tpr, color=colors[j],
                         label=r'%s (%0.2f $\pm$ %0.2f)' % (classes[j].strip("SN "), mean_auc, std_auc),
                         lw=2, alpha=.8)
        std_tpr = np.std(tprs, axis=0)
        tprs_upper = np.minimum(mean_tpr + std_tpr, 1)
        tprs_lower = np.maximum(mean_tpr - std_tpr, 0)
        ax.fill_between(mean_fpr, tprs_lower, tprs_upper, color=colors[j], alpha=.05)

    ax.plot([0, 1], [0, 1], linestyle='--', lw=2, color='k',alpha=.8)
    #if ~foley:
    ax.set_xlabel("False Positive Rate", fontsize=16);
    ax.set_ylabel("True Positive Rate", fontsize=16);
        #plt.title("ROC Curve, %i Classes" % (len(classes)), fontsize=26)
    ax.legend(loc=4,fontsize=12)
    #    plt.text(0.1, 0.9, r'$N_{tot} = %i$'%len(y),fontsize=12)
    plt.text(0.1, 0.9, r'$N_{train} = 7000$')
    plt.text(0.1, 0.82, r'$N_{test} = 2226$')
    ax.set_xlim([-0.05, 1.05])
    ax.set_ylim([-0.05, 1.05])
        #plt.savefig("Combined_MeanROC_Curve_%i_Classes_dataML_noOverlapCuts.png" % len(classes))
    return accuracy, rf, all_confMatrices, allAcc, wrong


def plot_ROC(train_features, test_features, train_labels, test_labels, save):
    rf = RandomForestClassifier(n_estimators = 1000, bootstrap = True, max_features = 'sqrt')
    rf.fit(train_features, train_labels);

    predictions = rf.predict(test_features)# Calculate the absolute errors

    fpr = dict()
    tpr = dict()
    ROC = dict()
    classes = np.unique(test_labels)
    plt.figure(figsize=(10,10))
    for i in range(len(classes)):
        rf_probs = rf.predict_proba(test_features)[:, i]
        fpr[i], tpr[i], _ = roc_curve(test_labels, rf_probs, pos_label=classes[i])
        ROC[i] = auc(fpr[i], tpr[i])
        plt.plot(fpr[i], tpr[i],
                 lw=4, label='%s (area = %0.2f)' % (classes[i], ROC[i]))
    accuracy = np.sum(predictions == test_labels)/len(test_labels)*100
    plt.plot([0, 1], [0, 1], color='navy', lw=2, linestyle='--')
    plt.legend(loc=4, fontsize=10)
    plt.xlabel("False Positive Rate", fontsize=16);
    plt.ylabel("True Positive Rate", fontsize=16);
    plt.title("ROC Curve, %i Classes, Accuracy = %.1f%%" % (len(classes), accuracy), fontsize=26)
    #os.chdir('/Users/alexgagliano/Documents/Research/Transient_ML/plots')
    if save:
        plt.savefig("ROC_Curve_%i_Classes_dataML.png" % len(classes))
    else:
        plt.show()
    return rf, predictions

#dataML = pd.read_csv("../database/GHOST.csv")

def condense_labels(dataML, nclass):
        # Labels are the values we want to predict
        dataML.loc[dataML['TransientClass'] == 'SN Ib\n SN Ib', 'TransientClass'] = 'SN Ib'
        dataML.loc[dataML['TransientClass'] == 'SN Ia\n SN Ia', 'TransientClass'] = 'SN Ia'
        dataML.loc[dataML['TransientClass'] == 'SN Ibn', 'TransientClass'] = 'SN Ib'
        dataML.loc[dataML['TransientClass'] == 'SN Ib', 'TransientClass'] = 'SN Ib/c'
        dataML.loc[dataML['TransientClass'] == 'SN Ic', 'TransientClass'] = 'SN Ib/c'
        dataML.loc[dataML['TransientClass'] == 'SLSN-I', 'TransientClass'] = 'SLSN'
        dataML.loc[dataML['TransientClass'] == 'SN I', 'TransientClass'] = 'SN I?'
        dataML.loc[dataML['TransientClass'] == 'SN Ib', 'TransientClass'] = 'SN Ib/c'
        dataML.loc[dataML['TransientClass'] == 'SN Ic', 'TransientClass'] = 'SN Ib/c'
        dataML.loc[dataML['TransientClass'] == 'SLSN-II', 'TransientClass'] = 'SLSN'
        dataML.loc[dataML['TransientClass'] == 'SLSN-II', 'TransientClass'] = 'SLSN'
        dataML.loc[dataML['TransientClass'] == 'CC', 'TransientClass'] = 'SN II'
        dataML.loc[dataML['TransientClass'] == 'II', 'TransientClass'] = 'SN II'
        dataML.loc[dataML['TransientClass'] == 'SLSN-I-R', 'TransientClass'] = 'SLSN'
        dataML.loc[dataML['TransientClass'] == 'SLSN-R', 'TransientClass'] = 'SLSN'
        dataML.loc[dataML['TransientClass'] == 'II/IIb', 'TransientClass'] = 'SN II'
        dataML.loc[dataML['TransientClass'] == 'II P', 'TransientClass'] = 'SN IIP'
        dataML.loc[dataML['TransientClass'] == 'Ib', 'TransientClass'] = 'SN Ib/c'
        dataML.loc[dataML['TransientClass'] == 'Ic', 'TransientClass'] = 'SN Ib/c'
        dataML.loc[dataML['TransientClass'] == 'II-p', 'TransientClass'] = 'SN II P'
        dataML.loc[dataML['TransientClass'] == 'II/LBV', 'TransientClass'] = 'SN II'
        dataML.loc[dataML['TransientClass'] == 'IIb', 'TransientClass'] = 'SN IIb'
        dataML.loc[dataML['TransientClass'] == 'Ic BL', 'TransientClass'] = 'SN Ib/c'
        dataML.loc[dataML['TransientClass'] == 'Ia', 'TransientClass'] = 'SN Ia'
        dataML.loc[dataML['TransientClass'] == 'Ib/c', 'TransientClass'] = 'SN Ib/c'
        dataML.loc[dataML['TransientClass'] == 'Ib/c', 'TransientClass'] = 'SN Ib/c'
        dataML.loc[dataML['TransientClass'] == 'IIn', 'TransientClass'] = 'SN IIn'
        dataML.loc[dataML['TransientClass'] == 'Ibn', 'TransientClass'] = 'SN Ib/c'
        dataML.loc[dataML['TransientClass'] == 'IIn Pec', 'TransientClass'] = 'SN IIn'
        dataML.loc[dataML['TransientClass'] == 'Ia/Ic', 'TransientClass'] = 'SN Ia/c'
        dataML.loc[dataML['TransientClass'] == 'SN II P', 'TransientClass'] = 'SN IIP'
        dataML.loc[dataML['TransientClass'] == 'Ia/c', 'TransientClass'] = 'SN Ia/c'
        dataML.loc[dataML['TransientClass'] == 'I', 'TransientClass'] = 'SN I'

        dataML.loc[dataML['TransientClass'] == 'LRV?', 'TransientClass'] = 'Unknown'
        dataML.loc[dataML['TransientClass'] == 'II Pec?', 'TransientClass'] = 'Unknown'
        dataML.loc[dataML['TransientClass'] == 'I?', 'TransientClass'] = 'Unknown'
        dataML.loc[dataML['TransientClass'] == 'SLSN-II?', 'TransientClass'] = 'Unknown'
        dataML.loc[dataML['TransientClass'] == 'IIb/Ib', 'TransientClass'] = 'Unknown'
        dataML.loc[dataML['TransientClass'] == 'Ib/Ic (Ca rich?)?', 'TransientClass'] = 'Unknown'
        dataML.loc[dataML['TransientClass'] == 'Ic?', 'TransientClass'] = 'Unknown'
        dataML.loc[dataML['TransientClass'] == 'IIn?', 'TransientClass'] = 'Unknown'
        dataML.loc[dataML['TransientClass'] == 'PISN?', 'TransientClass'] = 'Unknown'
        dataML.loc[dataML['TransientClass'] == 'SLSN?', 'TransientClass'] = 'Unknown'
        dataML.loc[dataML['TransientClass'] == 'Ib/c?', 'TransientClass'] = 'Unknown'
        dataML.loc[dataML['TransientClass'] == 'II?', 'TransientClass'] = 'Unknown'
        dataML.loc[dataML['TransientClass'] == 'IIb?', 'TransientClass'] = 'Unknown'
        dataML.loc[dataML['TransientClass'] == 'Ia?', 'TransientClass'] = 'Unknown'
        dataML.loc[dataML['TransientClass'] == 'SN I?', 'TransientClass'] = 'Unknown'
        dataML.loc[dataML['TransientClass'] == 'Ii', 'TransientClass'] = 'Unknown'
        dataML.loc[dataML['TransientClass'] == 'LBV to IIn', 'TransientClass'] = 'LBV'
        dataML.loc[dataML['TransientClass'] == 'II/Ib/c', 'TransientClass'] = 'Unknown'
        dataML.loc[dataML['TransientClass'] == 'Ca-rich', 'TransientClass'] = 'Unknown'
        dataML.loc[dataML['TransientClass'] == 'SN', 'TransientClass'] = 'Unknown'
        dataML.loc[dataML['TransientClass'] == 'SN Ia/c', 'TransientClass'] = 'Unknown'
        dataML.loc[dataML['TransientClass'] == 'Ib/IIb', 'TransientClass'] = 'Unknown'
        dataML.loc[dataML['TransientClass'] == 'IIn/LBV', 'TransientClass'] = 'IIn'
        dataML.loc[dataML['TransientClass'] == 'IIn', 'TransientClass'] = 'SN IIn'
        dataML.loc[dataML['TransientClass'] == 'IIn/LBV', 'TransientClass'] = 'IIn'
        dataML.loc[dataML['TransientClass'] == 'CV', 'TransientClass'] = 'Other'
        dataML.loc[dataML['TransientClass'] == 'Pec', 'TransientClass'] = 'Other'
        dataML.loc[dataML['TransientClass'] == 'LBV', 'TransientClass'] = 'Other'
        dataML.loc[dataML['TransientClass'] == 'IIn/LBV', 'TransientClass'] = 'IIn'
        dataML.loc[dataML['TransientClass'] == 'Ic Pec', 'TransientClass'] = 'Other'
        dataML.loc[dataML['TransientClass'] == 'CN', 'TransientClass'] = 'Other'
        dataML.loc[dataML['TransientClass'] == 'Ib-Ca', 'TransientClass'] = 'SN Ib/c'
        dataML.loc[dataML['TransientClass'] == 'SLSN-I?', 'TransientClass'] = 'SLSN'
        dataML.loc[dataML['TransientClass'] == 'SLSN-IIn', 'TransientClass'] = 'SLSN'
        dataML.loc[dataML['TransientClass'] == 'nIa', 'TransientClass'] = 'Unknown'
        dataML.loc[dataML['TransientClass'] == 'II L?', 'TransientClass'] = 'Unknown'
        dataML.loc[dataML['TransientClass'] == 'Ia-09dc', 'TransientClass'] = 'SN Ia'
        dataML.loc[dataML['TransientClass'] == 'CC?', 'TransientClass'] = 'Unknown'
        dataML.loc[dataML['TransientClass'] == 'SN Ia-pec', 'TransientClass'] = 'SN Ia Pec'#
        dataML.loc[dataML['TransientClass'] == 'SN Ia-91T-like', 'TransientClass'] = 'SN Ia Pec'#
        dataML.loc[dataML['TransientClass'] == 'SN Iax[02cx-like]', 'TransientClass'] = 'SN Ia Pec'#
        dataML.loc[dataML['TransientClass'] == 'SN Ia-91bg-like', 'TransientClass'] = 'SN Ia Pec'#
        dataML.loc[dataML['TransientClass'] == 'SN Ia-CSM', 'TransientClass'] = 'SN Ia Pec'#
        dataML.loc[dataML['TransientClass'] == 'SN Ia-91bg', 'TransientClass'] = 'SN Ia Pec'#
        dataML.loc[dataML['TransientClass'] == 'Ia Pec', 'TransientClass'] = 'SN Ia Pec'#
        dataML.loc[dataML['TransientClass'] == 'Ia*', 'TransientClass'] = 'SN Ia Pec'#
        dataML.loc[dataML['TransientClass'] == 'Ia-02cx', 'TransientClass'] = 'SN Ia Pec'#
        dataML.loc[dataML['TransientClass'] == 'Ia-91T', 'TransientClass'] = 'SN Ia Pec'#
        dataML.loc[dataML['TransientClass'] == 'Ia-91bg', 'TransientClass'] = 'SN Ia Pec'#
        dataML.loc[dataML['TransientClass'] == 'Ia-99aa', 'TransientClass'] = 'SN Ia Pec'#
        dataML.loc[dataML['TransientClass'] == 'Ia CSM', 'TransientClass'] = 'SN Ia Pec'#
        if nclass != 4:
            dataML = dataML[dataML['TransientClass'] != 'SN Ia Pec']

        # specific to the four-class
        if nclass == 5:
            dataML = dataML[dataML['TransientClass'] != 'SLSN']
            #dataML.loc[dataML['TransientClass'] == 'SN IIb', 'TransientClass'] = 'SN II'
            dataML = dataML[dataML['TransientClass'] != 'SN IIb']
            #dataML.loc[dataML['TransientClass'] == 'SN IIP', 'TransientClass'] = 'SN II'
            #dataML.loc[dataML['TransientClass'] == 'SN IIn', 'TransientClass'] = 'SN II'
            #II, IIP, Ia, IIn, Ib/c,
        # specific to the two-class
        elif nclass == 4:
            dataML.loc[dataML['TransientClass'] == 'SN II', 'TransientClass'] = 'Core Collapse'
            dataML.loc[dataML['TransientClass'] == 'SN IIP', 'TransientClass'] = 'Core Collapse'
            dataML.loc[dataML['TransientClass'] == 'SN IIb', 'TransientClass'] = 'Core Collapse'
            dataML.loc[dataML['TransientClass'] == 'SN IIn', 'TransientClass'] = 'Core Collapse'
            dataML.loc[dataML['TransientClass'] == 'SN Ib/c', 'TransientClass'] = 'Core Collapse'#
            #dataML = dataML[dataML['TransientClass'] != 'SN IIb']
            #dataML = dataML[dataML['TransientClass'] != 'SN Ib/c']
            #dataML = dataML[dataML['TransientClass'] != 'SN IIP']
            #dataML = dataML[dataML['TransientClass'] != 'SN IIn']
        elif nclass == 3:
            dataML.loc[dataML['TransientClass'] == 'SN II', 'TransientClass'] = 'Core Collapse'
            dataML.loc[dataML['TransientClass'] == 'SN IIP', 'TransientClass'] = 'Core Collapse'
            dataML.loc[dataML['TransientClass'] == 'SN IIb', 'TransientClass'] = 'Core Collapse'
            dataML.loc[dataML['TransientClass'] == 'SN IIn', 'TransientClass'] = 'Core Collapse'
            dataML.loc[dataML['TransientClass'] == 'SN Ib/c', 'TransientClass'] = 'Core Collapse'#

            #dataML = dataML[dataML['TransientClass'] != 'SN Ia']
            #dataML = dataML[dataML['TransientClass'] != 'Core Collapse']

            #SN II, SN Ib/c, SN Ia
        elif nclass == 2:
            dataML = dataML[dataML['TransientClass'] != 'SLSN']
            dataML.loc[dataML['TransientClass'] == 'SN II', 'TransientClass'] = 'Core Collapse'
            dataML.loc[dataML['TransientClass'] == 'SN IIP', 'TransientClass'] = 'Core Collapse'
            dataML.loc[dataML['TransientClass'] == 'SN IIb', 'TransientClass'] = 'Core Collapse'
            dataML.loc[dataML['TransientClass'] == 'SN IIn', 'TransientClass'] = 'Core Collapse'
            dataML.loc[dataML['TransientClass'] == 'SN Ib/c', 'TransientClass'] = 'Core Collapse'#

        dataML = dataML[dataML['TransientClass'] != 'SN II-pec']
        dataML = dataML[dataML['TransientClass'] != 'SLSN-I']
        dataML = dataML[dataML['TransientClass'] != 'SN Ic-BL']
        dataML = dataML[dataML['TransientClass'] != 'SN Ib-pec']
        dataML = dataML[dataML['TransientClass'] != 'SN Ib-Ca-rich']
        dataML = dataML[dataML['TransientClass'] != 'SN Ic-pec']
        dataML = dataML[dataML['TransientClass'] != 'SN IIL']
        dataML = dataML[dataML['TransientClass'] != 'II Pec']
        dataML = dataML[dataML['TransientClass'] != 'II L']
        dataML = dataML[dataML['TransientClass'] != 'Ib Pec']
        dataML = dataML[dataML['TransientClass'] != 'SN I']
        dataML = dataML[dataML['TransientClass'] != 'Other']
        dataML = dataML[dataML['TransientClass'] != 'Unknown']
        dataML = dataML[dataML['TransientClass'] != 'Ib-IIb']
        dataML = dataML[dataML['TransientClass'] != 'SNSN-II']

        return dataML

def preprocess_dataframe(dataML, nclass, PCA=False):
    dataML.replace(-999, np.nan, inplace=True)
    trueIDXs = dataML.dropna(subset=['TransientRedshift', 'NED_redshift']).index
    naIDXs = set(dataML.index) - set(trueIDXs)

    dataML_na = dataML.loc[naIDXs]
    dataML_nona = dataML.loc[trueIDXs]

    pdiff = np.abs(dataML_nona['TransientRedshift'] - dataML_nona['NED_redshift'])/dataML_nona['TransientRedshift']*100
    dataML_nona = dataML_nona.loc[pdiff < 5]
    dataML = pd.concat([dataML_na, dataML_nona], ignore_index=True)

    #ADDING IN SNR IN TWO BANDS - I AND Z
    dataML["gSNR"] = 1/dataML["gApMagErr"]
    dataML["rSNR"] = 1/dataML["rApMagErr"]
    dataML["iSNR"] = 1/dataML["iApMagErr"]
    dataML["zSNR"] = 1/dataML["zApMagErr"]
    dataML["ySNR"] = 1/dataML["yApMagErr"]

    dataML = dataML.drop(['objAltName1', 'objAltName2','objAltName3'], axis=1)
    dataML = dataML.drop(['objName','uniquePspsOBid','ippObjID','surveyID','htmID','zoneID','tessID','projectionID','skyCellID'], axis=1)
    dataML = dataML.drop(['randomID','batchID','dvoRegionID','processingVersion','objInfoFlag','qualityFlag','raStack','decStack'], axis=1)
    dataML = dataML.drop(['raStackErr', 'decStackErr', 'raMean', 'decMean', 'raMeanErr', 'decMeanErr'], axis=1)
    dataML = dataML.drop(['gra', 'gdec', 'graErr', 'gdecErr', 'rra', 'rdec', 'rraErr', 'rdecErr','ira', 'idec', 'iraErr', 'idecErr','zra', 'zdec', 'zraErr', 'zdecErr','yra', 'ydec', 'yraErr', 'ydecErr'], axis=1)
    dataML = dataML.drop(['l','b','nStackObjectRows'],axis=1)
    dataML = dataML.drop(['nStackDetections','nDetections'],axis=1)
    dataML = dataML.drop(['gippDetectID', 'gstackDetectID', 'gstackImageID','rippDetectID', 'rstackDetectID', 'rstackImageID','iippDetectID', 'istackDetectID', 'istackImageID','zippDetectID', 'zstackDetectID', 'zstackImageID','yippDetectID', 'ystackDetectID', 'ystackImageID'], axis=1)
    dataML = dataML.drop(['bestDetection'],axis=1)
    dataML = dataML.drop(['epochMean'],axis=1)
    dataML = dataML.drop(['ng','nr','ni','nz'],axis=1)
    dataML = dataML.drop(['ny'],axis=1)
    dataML = dataML.drop(['uniquePspsSTid','primaryDetection','gEpoch'],axis=1)
    dataML = dataML.drop(['rEpoch','iEpoch','zEpoch', 'yEpoch'],axis=1)
    dataML = dataML.drop(['cx','cy'],axis=1)
    dataML = dataML.drop(['cz'],axis=1)
    dataML = dataML.drop(['lambda','beta'],axis=1)
    dataML = dataML.drop(['gpsfChiSq','rpsfChiSq','ipsfChiSq','zpsfChiSq','ypsfChiSq', 'ginfoFlag', 'ginfoFlag2', 'ginfoFlag3',  'rinfoFlag', 'rinfoFlag2', 'rinfoFlag3',  'iinfoFlag', 'iinfoFlag2', 'iinfoFlag3',  'zinfoFlag', 'zinfoFlag2', 'zinfoFlag3',  'yinfoFlag', 'yinfoFlag2', 'yinfoFlag3'],axis=1)
    dataML = dataML.drop(['gxPos', 'gxPosErr','rxPos', 'rxPosErr','ixPos', 'ixPosErr','zxPos', 'zxPosErr','yxPos', 'yxPosErr' ],axis=1)
    dataML = dataML.drop(['gyPos', 'gyPosErr','ryPos', 'ryPosErr','iyPos', 'iyPosErr','zyPos', 'zyPosErr','yyPos', 'yyPosErr' ],axis=1)
    dataML = dataML.drop(['gexpTime','rexpTime','iexpTime','zexpTime','yexpTime','gnFrames','rnFrames','inFrames','znFrames','ynFrames'],axis=1)
    dataML = dataML.drop(['gzp','rzp','izp','zzp','yzp'],axis=1)
    dataML = dataML.drop(['gPlateScale','rPlateScale','iPlateScale','zPlateScale','yPlateScale'],axis=1)
    dataML = dataML.drop(['posMeanChisq'],axis=1)
    dataML = dataML.drop(['gpsfQf','ipsfQf', 'zpsfQf', 'ypsfQf'], axis=1)
    dataML = dataML.drop(['gApFillFac', 'yApFillFac', 'iApFillFac', 'zApFillFac'], axis=1)
    dataML = dataML.drop(['gpsfQfPerfect', 'ipsfQfPerfect', 'zpsfQfPerfect', 'ypsfQfPerfect'], axis=1)
    dataML = dataML.drop(['gpsfTheta', 'ipsfTheta', 'zpsfTheta', 'ypsfTheta'], axis=1)
    dataML = dataML.drop(['gsky', 'isky', 'zsky', 'ysky'], axis=1)
    dataML = dataML.drop(['gskyErr', 'iskyErr', 'zskyErr', 'yskyErr'], axis=1)
    dataML = dataML.drop(['gpsfCore', 'ipsfCore', 'zpsfCore', 'ypsfCore'], axis=1)
    dataML = dataML.drop(['rpsfTheta', 'rsky', 'rskyErr', 'rpsfCore'], axis=1)
    dataML = dataML.drop(['gpsfLikelihood', 'rpsfLikelihood', 'ipsfLikelihood', 'zpsfLikelihood','ypsfLikelihood'], axis=1)
    dataML = dataML.drop(['rpsfQf'], axis=1)
    dataML = dataML.drop(['host_logmass', 'host_logmass_min', 'host_logmass_max','Hubble Residual', 'Transient AltName'],axis=1)
    dataML = dataML.drop(['rpsfQfPerfect'], axis=1)
    dataML = dataML.drop(['rApFillFac'], axis=1)
    dataML = dataML.drop(['TransientRA', 'TransientDEC','NED_type', 'NED_name'], axis=1)

    #try dropping NED info now:
    dataML = dataML.drop(['NED_vel', 'NED_mag'], axis=1)
    dataML = dataML.drop(['NED_redshift'], axis=1)
    dataML = dataML.drop(['TransientRedshift'], axis=1)
    dataML.drop(['TransientDiscoveryDate',  'TransientDiscoveryMag', 'TransientDiscoveryYear'], axis=1, inplace=True)
    dataML = dataML.drop(['objID'],axis=1)

    if PCA == True:
        allCols = []
        bands = ['g', 'r', 'i', 'z', 'y']
        cols = ['KronMag', 'ApMag', 'PSFMag', 'PSFFlux', 'ApFlux', 'KronFlux', 'KronRad', 'momentR1', 'ApMag_KronMag', 'momentXX', 'momentYY', 'momentRH', 'ExtNSigma']
        for band in bands:
            for col in cols:
                if col == 'ApMag_KronMag':
                    allCols.append(band+ col.split("_")[0] + "_"+ band + col.split("_")[1])
                else:
                    allCols.append(band+col)
        allCols.append("TransientClass")
        allCols.append("TransientName")
        dataML = dataML[allCols]
    dataML = dataML.dropna()
    print(dataML.shape)
    #dataML

    dataML = condense_labels(dataML, nclass=nclass)
    names = dataML['TransientName']
    dataML = dataML.drop(['TransientName'], axis=1)
    labels_df = dataML['TransientClass']# Remove the labels from the features
    labels = np.array(labels_df)
    classes = np.unique(labels)

    feature_list = list(dataML.columns) # Convert to numpy array

    dataML_noLabels = dataML.drop('TransientClass', axis=1)

    print('Distribution before imbalancing: {}'.format(Counter(labels)))

    labels_df.reset_index(inplace=True, drop=True)
    return feature_list, dataML_noLabels, labels_df, names

#feature_list, dataML_preprocessed2, labels_df2, names = preprocess_dataframe(dataML, nclass=2, PCA=False)
#importances2class = get_importances(dataML_preprocessed2, labels_df2, nclass=2, save=0)
##transform the data
#dataML_matrix_scaled = preprocessing.scale(dataML_preprocessed2)
#labels = labels_df2.values

#labels_df2.reset_index(inplace=True, drop=True)
#fig2 = plt.figure(figsize=(5.0, 4.0), dpi=300) #frameon=false
#ax = fig2.gca()
#acc, rf, all_confMatrices, accTot, wrong = plot_ROC_wCV(ax, dataML_matrix_scaled, labels.ravel(), names.values, save=0, balance=True)

#fs.plot_feature_importances(threshold = 0.99, plot_n = 12)

########################################################################################
########################################################################################

##
##                   CODE BELOW FOR CALCULATING VARIABLE IMPORTANCES
##


########################################################################################
#######################################################################################

#importances2class['class'] = '2 class'
#importances4class['class'] = '4 class'
#importances7class['class'] = '7 class'

#importances4class.sort_values(by=['normalized_importance'], ascending=False, inplace=True)
#importances4class_topTen = importances4class[importances4class['feature'].isin(['g-r', 'dist', 'dist/DLR', 'g-i', 'r-i', 'gExtNSigma', 'i-z', '7DCD', 'z-y', 'iApMag_iKronMag'])]

#importances2class.sort_values(by=['normalized_importance'], ascending=False, inplace=True)
#print(importances2class['feature'].values)

#importances2class_topTen = importances2class[importances2class['feature'].isin(['dist', 'dist/DLR', 'g-r', 'g-i', 'r-i', 'i-z', '7DCD','gExtNSigma', 'imomentXX', 'gmomentXX'])]

#importances7class.sort_values(by=['normalized_importance'], ascending=False, inplace=True)
#importances7class_topTen = importances7class[importances7class['feature'].isin(['g-r', 'dist', 'dist/DLR', 'g-i', 'r-i', 'gExtNSigma', 'i-z', '7DCD', 'z-y', 'iApMag_iKronMag'])]

#importances = importances2class_topTen.append(importances4class_topTen, ignore_index=True)
#importances_final = importances.append(importances7class_topTen, ignore_index=True)
#importances_final = importances2class_topTen
#importances_final.loc[importances_final['feature'] == 'dist', 'feature'] = r'$\theta$'
#importances_final.loc[importances_final['feature'] == 'dist/DLR', 'feature'] = r'$\theta/d_{DLR}$'

#importances_final.loc[importances_final['feature'] == 'g-r', 'feature'] = r'$g-r$'
#importances_final.loc[importances_final['feature'] == 'r-i', 'feature'] = r'$r-i$'
#importances_final.loc[importances_final['feature'] == 'g-i', 'feature'] = r'$g-i$'
#importances_final.loc[importances_final['feature'] == '7DCD', 'feature'] = '4DCD'
#importances_final.loc[importances_final['feature'] == 'i-z', 'feature'] = r'$i-z$'
#importances_final.loc[importances_final['feature'] == 'z-y', 'feature'] = r'$z-y$'
#importances_final.loc[importances_final['feature'] == 'iApMag_iKronMag', 'feature'] = r'$m_{\rm Ap, i} - m_{\rm Kron, i}$'
#importances_final.loc[importances_final['feature'] == 'gExtNSigma', 'feature'] = r'ExtNSigma, $g$'
#importances_final.loc[importances_final['feature'] == 'gmomentXX', 'feature'] = r'momentXX, $g$'
#importances_final.loc[importances_final['feature'] == 'imomentXX', 'feature'] = r'momentXX, $i$'

#importances_final['Importance'] = importances_final['normalized_importance']
#importances_final['Feature'] = importances_final['feature']
#importances_final['Model'] = importances_final['class']

#sns.set_context("talk")
#ax = sns.barplot(y="Feature", x="Importance", data=importances_final, palette="Blues_d")
#plt.legend(fontsize=16)
#ax.set_ylabel("Feature",fontsize=22)
#ax.set_xlabel("Importance",fontsize=22)
#plt.savefig("FeatureImportances_0715.png", dpi=300, bbox_inches='tight')



from bs4 import BeautifulSoup
import pandas as pd
import requests
import os
from astro_ghost.PS1QueryFunctions import create_df
import urllib.request
from os import listdir
import numpy as np
from os.path import isfile, join
import matplotlib.pyplot as plt
from astro_ghost.PS1QueryFunctions import find_all

def remove_comments(line,sep="#"):
    for s in sep:
        i = line.find(s)#find the position of
        if i == 0 :
            line = None
    return line

def clean_spectra(dir):
    for filename in os.listdir(dir):
        try:
            data = pd.read_csv(dir+'/' +filename, delim_whitespace=True, header=None)
            plt.plot(data[0], data[1])
            plt.clf()
        except:
            print(filename)
            with open(dir+filename, "r+") as f:
                lines = f.readlines()
                f.seek(0)
                for line in lines:
                    if not line.startswith("#"):
                        f.write(line)
                    else:
                        print(line)
                f.truncate()

def remove_prefix(text, prefix):
    return text[text.startswith(prefix) and len(prefix):]

# combine OSC and TNS transients!
# scrape new TNS objects and
#
def get_new_transients(period, period_unit):
    url = "https://wis-tns.weizmann.ac.il/search?&discovered_period_value=%i&discovered_period_units=%s&unclassified_at=0&classified_sne=0&name=&name_like=0&isTNS_AT=all&public=all&ra=&decl=&radius=&coords_unit=arcsec&reporting_groupid%5B%5D=null&groupid%5B%5D=null&classifier_groupid%5B%5D=null&objtype%5B%5D=18%2C19%2C20%2C1%2C2%2C15%2C16%2C3%2C103%2C104%2C106%2C100%2C102%2C105%2C4%2C8%2C107%2C6%2C9%2C5%2C7%2C108%2C10%2C110%2C14%2C12%2C13%2C112%2C11&at_type%5B%5D=null&date_start%5Bdate%5D=&date_end%5Bdate%5D=&discovery_mag_min=&discovery_mag_max=&internal_name=&discoverer=&classifier=&spectra_count=&redshift_min=&redshift_max=&hostname=&ext_catid=&ra_range_min=&ra_range_max=&decl_range_min=&decl_range_max=&discovery_instrument%5B%5D=null&classification_instrument%5B%5D=null&associated_groups%5B%5D=null&at_rep_remarks=&class_rep_remarks=&num_page=1000" %(period, period_unit)
    return df

##will have the website scraping
def get_transients(table_dir, fn_SN, tns_loc):
    # current date and time
    #but down the line - write a script to Scrape all new SNe
    #create a transient dataframe from TNS queries
    transient_df = create_df(tns_loc)
    transient_df = transient_df.drop_duplicates()
    transient_df.to_csv(table_dir + fn_SN)

#AND TNS spectrum scraping
def getTNSSpectra(transients, path, verbose=0):
    names = [remove_prefix(x, "SN") for x in transients['Name']]
    for name in names:
        URL = 'https://wis-tns.weizmann.ac.il/object/' + name
        page = requests.get(URL)
        soup = BeautifulSoup(page.content, 'html.parser')
        files = soup.findAll("td", {"class": "cell-asciifile"})

        if (len(files) > 0):
            if verbose:
                print("%s found in TNS. Downloading spectra:"%name)
            for file in files:
                link = file.a['href']
                fn = link.split("/")[-1]
                a = find_all(fn, path)
                if not a:
                    urllib.request.urlretrieve(link, fn)
        else:
            if verbose:
                print("No spectra for %s found on TNS."%name)
            else:
                continue



from sklearn import svm
import seaborn as sns
#from matplotlib.pyplot import figure
import matplotlib.pyplot as plt
from astro_ghost.stellarLocus import *
import os
import pickle
import pkg_resources

#10/11/2020: a new star/galaxy separation model, this time using a random forest classifier saved with hyperparameters chosen with randomGridSearch
def separateStars_STRM(df, model_path='.', plot=0, verbose=0, starcut='gentle'):
    df_dropped = df.dropna(subset=['7DCD','gApMag','gApMag_gKronMag','rApMag','rApMag_rKronMag','iApMag', 'iApMag_iKronMag'])
    only_na = df[~df.index.isin(df_dropped.index)]
    unsure = df_dropped.reset_index(drop=True)

    modelName = "Star_Galaxy_RealisticModel_GHOST_PS1ClassLabels.sav"
    stream = pkg_resources.resource_stream(__name__, modelName)
    if verbose:
        print("Loading model %s."%modelName)
    model = pickle.load(stream)

    if plot:
        sns.set_style("dark")
        sns.set_context("talk")

        plt.figure(figsize=(10,8))
        plt.plot(unsure['iApMag'], unsure['iApMag_iKronMag'], 'o', alpha=0.1)
        plt.xlabel(r"Ap Mag, $i$")
        plt.ylabel(r"Ap - Kron Mag, $i$")
        plt.savefig("TNS_NoClassificationInfo.pdf")

    #NED_stars_nona = NED_stars.dropna(subset=['iApMag','iApMag_iKronMag'])
    #NED_gals_nona = NED_gals.dropna(subset=['iApMag','iApMag_iKronMag'])

    unsure = unsure.dropna(subset=['7DCD','gApMag','gApMag_gKronMag','rApMag','rApMag_rKronMag','iApMag', 'iApMag_iKronMag'])
    if len(unsure) <1:
        if verbose:
            print("No sources in field with feature values, skipping star/galaxy separation...")
        return df, unsure
    test_X = np.matrix(unsure[['7DCD','gApMag','gApMag_gKronMag','rApMag','rApMag_rKronMag','iApMag', 'iApMag_iKronMag']])
    if starcut is 'normal':
        test_y = model.predict(test_X)
    elif starcut is 'aggressive':
        test_y = model.predict_proba(test_X)[:,1] > 0.3
    elif starcut is 'gentle':
        test_y = model.predict_proba(test_X)[:,1] > 0.8
    else:
        print("Error! I didn't understand your starcut option.")
    unsure['class'] = test_y
    test_stars = unsure[unsure['class'] == 1]
    test_gals = unsure[unsure['class'] == 0]

    if plot:
        c_gals = '#087e8b'
        c_stars = '#ffbf00'

        plt.figure(num=None, figsize=(8, 6), dpi=80, facecolor='w', edgecolor='k')
        sns.kdeplot(test_gals['iApMag'], test_gals['iApMag_iKronMag'], n_levels=20, shade_lowest=False, shade=False,label='Galaxies', color=c_gals, alpha=0.5,legend=True);
        if len(test_stars) > 1:
            sns.kdeplot(test_stars['iApMag'], test_stars['iApMag_iKronMag'], n_levels=20, shade_lowest=False, shade=False,label='Stars', color=c_stars, alpha=0.5);
        plt.legend()
        plt.xlim(xmin=13,xmax=23)
        plt.ylim(ymin=-1,ymax=5)
        plt.xlabel("Ap Magnitude, $i$",fontsize=16)
        plt.ylabel("Ap - Kron Magnitude, $i$",fontsize=16)
        #plt.legend(fontsize=16)
        plt.savefig("TNS_STRM_Stars_vs_Gals_RF_Contours.pdf")

        plt.figure(num=None, figsize=(8, 6), dpi=80, facecolor='w', edgecolor='k')
        ax = sns.distplot(test_gals['iApMag_iKronMag'],bins=np.linspace(-1, 7, num=100),label='Galaxies',kde=False,hist_kws={"color": c_gals});
        if len(test_stars) > 1:
            sns.distplot(test_stars['iApMag_iKronMag'],bins=np.linspace(-1, 7, num=100),label='Stars',kde=False,hist_kws={"color": c_stars});
        sns.distplot(unsure['iApMag_iKronMag'],bins=np.linspace(-1, 7, num=100),label='Total',kde=False,hist_kws={"histtype": "step", "linewidth": 1, "alpha": 1, "color": "black"});
        #ax.set_yscale('log')
        plt.xlim(-1, 3)
        plt.xlabel("Ap - Kron Mag, $i$",fontsize=16)
        plt.xlim()
        plt.ylabel("N",fontsize=16)
        plt.legend(fontsize=16)
        plt.savefig("TNS_iAp_iKronMag_Histogram_RF.pdf")

    # removing the ones where we have no NED identification
    df = df[~df['NED_type'].isnull()]

    # now we want to add back in the ones we identified as galaxies from our clustering above:
    df_gals = pd.concat([test_gals, only_na])
    df_gals.reset_index(inplace=True, drop=True)
    df_stars = test_stars.drop_duplicates()
    #df_stars.to_csv('OSC_061019_PS1_stars_109_NEDCuts.tar.gz')
    #df_gals.to_csv('TNS_PS1_gals_109_NEDCuts.tar.gz')
    return df_gals, df_stars#

#10/06/2020: a new star/galaxy separation model, this time using a random forest classifier saved with hyperparameters chosen with randomGridSearch
def separateStars_RF(df, model_path='.', plot=0, verbose=0):
    df.replace('', np.nan, inplace=True)
    df_dropped = df.dropna(subset=['7DCD','gApMag','gApMag_gKronMag','rApMag','rApMag_rKronMag','iApMag', 'iApMag_iKronMag'])
    only_na = df[~df.index.isin(df_dropped.index)]
    df = df_dropped.reset_index(drop=True)

    NED_stars = df[df['NED_type'] == '*']
    if verbose:
        print("Found %s NED-identified stars in the candidate list!"%len(NED_stars))
    NED_gals = df[df['NED_type'] == 'G']
    if verbose:
        print("Found %s NED-identified galaxies in the candidate list!"%len(NED_gals))
    NED_unsure= df[df['NED_type'].isnull()]

    modelName = "Star_Galaxy_RealisticModel.sav"
    stream = pkg_resources.resource_stream(__name__, modelName)
    model = pickle.load(stream)

    if plot:
        sns.set_style("dark")
        sns.set_context("talk")

        plt.figure(figsize=(10,8))
        plt.plot(NED_unsure['iApMag'], NED_unsure['iApMag_iKronMag'], 'o', alpha=0.1)
        plt.xlabel(r"Ap Mag, $i$")
        plt.ylabel(r"Ap - Kron Mag, $i$")
        plt.savefig("TNS_NoNedInfo.pdf")

    NED_stars_nona = NED_stars.dropna(subset=['iApMag','iApMag_iKronMag'])
    NED_gals_nona = NED_gals.dropna(subset=['iApMag','iApMag_iKronMag'])

    if plot:
        figure(num=None, figsize=(8, 6), dpi=80, facecolor='w', edgecolor='k')
        sns.kdeplot(NED_gals_nona['iApMag'], NED_gals_nona['iApMag_iKronMag'], n_levels=20, shade_lowest=False, shade=False,label='Galaxies', alpha=0.5,legend=True);
        sns.kdeplot(NED_stars_nona['iApMag'], NED_stars_nona['iApMag_iKronMag'], n_levels=20, shade_lowest=False, shade=False,label='Stars', alpha=0.5);
        plt.legend()
        plt.xlim(xmin=12,xmax=24)
        plt.ylim(ymin=-1,ymax=4)
        plt.xlabel("Ap Magnitude, $i$",fontsize=16)
        plt.ylabel("Ap - Kron Magnitude, $i$",fontsize=16)
        plt.savefig("TNS_NED_Stars_vs_Gals_Contours.pdf")

        plt.figure(figsize=(10,8))
        plt.plot(NED_gals['7DCD'], NED_gals['iApMag_iKronMag'],'o', alpha=0.1, color='k',label='NED Galaxies')
        plt.plot(NED_stars['7DCD'], NED_stars['iApMag_iKronMag'],'o', alpha=0.1, color='r', label='NED Stars')
        plt.xlabel(r"7DCD, $i$")
        plt.xscale("log")
        plt.ylabel(r"Ap - Kron Mag, $i$")
        plt.legend(fontsize=20)
        plt.savefig("TNS_NED_Stars_vs_Gals_v7DCD.pdf")

    NED_unsure = NED_unsure.dropna(subset=['7DCD','gApMag','gApMag_gKronMag','rApMag','rApMag_rKronMag','iApMag', 'iApMag_iKronMag'])
    NED_test_X = np.matrix(NED_unsure[['7DCD','gApMag','gApMag_gKronMag','rApMag','rApMag_rKronMag','iApMag', 'iApMag_iKronMag']])

    if verbose:
        print("Running star-galaxy separator on %s sources...."%len(NED_unsure))

    NED_test_y = model.predict(NED_test_X)
    NED_unsure['class'] = NED_test_y
    NED_test_stars = NED_unsure[NED_unsure['class'] == 1]
    NED_test_gals = NED_unsure[NED_unsure['class'] == 0]

    if plot:
        c_gals = '#087e8b'
        c_stars = '#ffbf00'

        figure(num=None, figsize=(8, 6), dpi=80, facecolor='w', edgecolor='k')
        sns.kdeplot(NED_test_gals['iApMag'], NED_test_gals['iApMag_iKronMag'], n_levels=20, shade_lowest=False, shade=False,label='Galaxies', color=c_gals, alpha=0.5,legend=True);
        if len(NED_test_stars) > 1:
            sns.kdeplot(NED_test_stars['iApMag'], NED_test_stars['iApMag_iKronMag'], n_levels=20, shade_lowest=False, shade=False,label='Stars', color=c_stars, alpha=0.5);
        plt.legend()
        plt.xlim(xmin=13,xmax=23)
        plt.ylim(ymin=-1,ymax=5)
        plt.xlabel("Ap Magnitude, $i$",fontsize=16)
        plt.ylabel("Ap - Kron Magnitude, $i$",fontsize=16)
        #plt.legend(fontsize=16)
        plt.savefig("TNS_NED_Stars_vs_Gals_SVM_Contours.pdf")

        figure(num=None, figsize=(8, 6), dpi=80, facecolor='w', edgecolor='k')
        ax = sns.distplot(NED_test_gals['iApMag_iKronMag'],bins=np.linspace(-1, 7, num=100),label='Galaxies',kde=False,hist_kws={"color": c_gals});
        if len(NED_test_stars) > 1:
            sns.distplot(NED_test_stars['iApMag_iKronMag'],bins=np.linspace(-1, 7, num=100),label='Stars',kde=False,hist_kws={"color": c_stars});
        sns.distplot(NED_unsure['iApMag_iKronMag'],bins=np.linspace(-1, 7, num=100),label='Total',kde=False,hist_kws={"histtype": "step", "linewidth": 1, "alpha": 1, "color": "black"});
        #ax.set_yscale('log')
        plt.xlim(-1, 3)
        plt.xlabel("Ap - Kron Mag, $i$",fontsize=16)
        plt.xlim()
        plt.ylabel("N",fontsize=16)
        plt.legend(fontsize=16)
        plt.savefig("TNS_iAp_iKronMag_Histogram_SVM_NEDInfo.pdf")

    # removing the ones where we have no NED identification
    df = df[~df['NED_type'].isnull()]

    # now we want to add back in the ones we identified as galaxies from our clustering above:
    df_gals = pd.concat([df, NED_test_gals, only_na])

    #removing some objects we know not to be our hosts
    df_gals = df_gals[df_gals['NED_type'] != 'HII']
    df_gals = df_gals[df_gals['NED_type'] != '*']
    df_gals = df_gals[df_gals['NED_type'] != '**']
    df_gals = df_gals[df_gals['NED_type'] != 'QSO']
    df_gals = df_gals[df_gals['NED_type'] != 'SNR']
    df_gals = df_gals[df_gals['NED_type'] != 'WD*']
    df_gals = df_gals[df_gals['NED_type'] != '!*']
    df_gals = df_gals[df_gals['NED_type'] != '!V*']
    df_gals = df_gals[df_gals['NED_type'] != 'V*']
    df_gals = df_gals[df_gals['NED_type'] != 'PN']
    df_gals = df_gals[df_gals['NED_type'] != '!Nova']

    #df = df[df['NED_type'] != 'PofG']

    df_gals.reset_index(inplace=True, drop=True)
    df_stars = pd.concat([NED_test_stars, NED_stars]).drop_duplicates()
    #df_stars.to_csv('OSC_061019_PS1_stars_109_NEDCuts.tar.gz')
    #df_gals.to_csv('TNS_PS1_gals_109_NEDCuts.tar.gz')
    return df_gals, df_stars#


def separateStars(df, plot=0):
    df_dropped = df.dropna(subset=['7DCD','gApMag','gApMag_gKronMag','rApMag','rApMag_rKronMag','iApMag', 'iApMag_iKronMag'])
    only_na = df[~df.index.isin(df_dropped.index)]
    df = df_dropped.reset_index(drop=True)

    NED_stars = df[df['NED_type'] == '*']
    NED_gals = df[df['NED_type'] == 'G']
    NED_unsure= df[df['NED_type'].isnull()]

    if plot:
        sns.set_style("dark")
        sns.set_context("talk")

        plt.figure(figsize=(10,8))
        plt.plot(NED_unsure['iApMag'], NED_unsure['iApMag_iKronMag'], 'o', alpha=0.1)
        plt.xlabel(r"Ap Mag, $i$")
        plt.ylabel(r"Ap - Kron Mag, $i$")
        plt.savefig("TNS_NoNedInfo.pdf")

    NED_stars['class'] = 1
    NED_stars = NED_stars[NED_stars['iApMag_iKronMag'] < 1.0]
    NED_gals['class'] = 0

    NED_stars_nona = NED_stars.dropna(subset=['iApMag','iApMag_iKronMag'])
    NED_gals_nona = NED_gals.dropna(subset=['iApMag','iApMag_iKronMag'])

    if plot:
        figure(num=None, figsize=(8, 6), dpi=80, facecolor='w', edgecolor='k')
        sns.kdeplot(NED_gals_nona['iApMag'], NED_gals_nona['iApMag_iKronMag'], n_levels=20, shade_lowest=False, shade=False,label='Galaxies', alpha=0.5,legend=True);
        sns.kdeplot(NED_stars_nona['iApMag'], NED_stars_nona['iApMag_iKronMag'], n_levels=20, shade_lowest=False, shade=False,label='Stars', alpha=0.5);
        plt.legend()
        plt.xlim(xmin=12,xmax=24)
        plt.ylim(ymin=-1,ymax=4)
        plt.xlabel("Ap Magnitude, $i$",fontsize=16)
        plt.ylabel("Ap - Kron Magnitude, $i$",fontsize=16)
        plt.savefig("TNS_NED_Stars_vs_Gals_Contours.pdf")

        plt.figure(figsize=(10,8))
        plt.plot(NED_gals['7DCD'], NED_gals['iApMag_iKronMag'],'o', alpha=0.1, color='k',label='NED Galaxies')
        plt.plot(NED_stars['7DCD'], NED_stars['iApMag_iKronMag'],'o', alpha=0.1, color='r', label='NED Stars')
        plt.xlabel(r"7DCD, $i$")
        plt.xscale("log")
        plt.ylabel(r"Ap - Kron Mag, $i$")
        plt.legend(fontsize=20)
        plt.savefig("TNS_NED_Stars_vs_Gals_v7DCD.pdf")

    # create an SVM model to predict our stars and our galaxies!
    clf = svm.SVC()

    gals_OSC = pd.read_csv("/Users/alexgagliano/Documents/Research/Transient_ML_Box/tables/SVM_training_gals.tar.gz")
    stars_OSC = pd.read_csv("/Users/alexgagliano/Documents/Research/Transient_ML_Box/tables/SVM_training_stars.tar.gz")

    if (len(NED_stars) > 500) and (len(NED_gals) > 500):
        # get rid of the outliers !
        NED_stars = NED_stars[NED_stars['iApMag_iKronMag'] < 1.0]
        NED_stars['class'] = 1
        NED_gals['class'] = 0
        NED_training = pd.concat([NED_stars, NED_gals])
    else:
        NED_training = pd.concat([stars_OSC,gals_OSC])

    NED_training = NED_training.dropna(subset=['7DCD','gApMag','gApMag_gKronMag','rApMag','rApMag_rKronMag','iApMag', 'iApMag_iKronMag'])
    NED_training_X = np.matrix(NED_training[['7DCD','gApMag','gApMag_gKronMag','rApMag','rApMag_rKronMag','iApMag', 'iApMag_iKronMag']])
    NED_training_y = np.array(NED_training['class'])
    clf.fit(NED_training_X, NED_training_y)

    NED_unsure = NED_unsure.dropna(subset=['7DCD','gApMag','gApMag_gKronMag','rApMag','rApMag_rKronMag','iApMag', 'iApMag_iKronMag'])
    NED_test_X = np.matrix(NED_unsure[['7DCD','gApMag','gApMag_gKronMag','rApMag','rApMag_rKronMag','iApMag', 'iApMag_iKronMag']])
    NED_test_y = clf.predict(NED_test_X)
    NED_unsure['class'] = NED_test_y
    NED_test_stars = NED_unsure[NED_unsure['class'] == 1]
    NED_test_gals = NED_unsure[NED_unsure['class'] == 0]

    if plot:
        c_gals = '#087e8b'
        c_stars = '#ffbf00'

        figure(num=None, figsize=(8, 6), dpi=80, facecolor='w', edgecolor='k')
        sns.kdeplot(NED_test_gals['iApMag'], NED_test_gals['iApMag_iKronMag'], n_levels=20, shade_lowest=False, shade=False,label='Galaxies', color=c_gals, alpha=0.5,legend=True);
        if len(NED_test_stars) > 1:
            sns.kdeplot(NED_test_stars['iApMag'], NED_test_stars['iApMag_iKronMag'], n_levels=20, shade_lowest=False, shade=False,label='Stars', color=c_stars, alpha=0.5);
        plt.legend()
        plt.xlim(xmin=13,xmax=23)
        plt.ylim(ymin=-1,ymax=5)
        plt.xlabel("Ap Magnitude, $i$",fontsize=16)
        plt.ylabel("Ap - Kron Magnitude, $i$",fontsize=16)
        #plt.legend(fontsize=16)
        plt.savefig("TNS_NED_Stars_vs_Gals_SVM_Contours.pdf")

        figure(num=None, figsize=(8, 6), dpi=80, facecolor='w', edgecolor='k')
        ax = sns.distplot(NED_test_gals['iApMag_iKronMag'],bins=np.linspace(-1, 7, num=100),label='Galaxies',kde=False,hist_kws={"color": c_gals});
        if len(NED_test_stars) > 1:
            sns.distplot(NED_test_stars['iApMag_iKronMag'],bins=np.linspace(-1, 7, num=100),label='Stars',kde=False,hist_kws={"color": c_stars});
        sns.distplot(NED_unsure['iApMag_iKronMag'],bins=np.linspace(-1, 7, num=100),label='Total',kde=False,hist_kws={"histtype": "step", "linewidth": 1, "alpha": 1, "color": "black"});
        #ax.set_yscale('log')
        plt.xlim(-1, 3)
        plt.xlabel("Ap - Kron Mag, $i$",fontsize=16)
        plt.xlim()
        plt.ylabel("N",fontsize=16)
        plt.legend(fontsize=16)
        plt.savefig("TNS_iAp_iKronMag_Histogram_SVM_NEDInfo.pdf")

    # removing the ones where we have no NED identification
    df = df[~df['NED_type'].isnull()]

    # now we want to add back in the ones we identified as galaxies from our clustering above:
    df_gals = pd.concat([df, NED_test_gals, only_na])

    #removing some objects we know not to be our hosts
    df_gals = df_gals[df_gals['NED_type'] != 'HII']
    df_gals = df_gals[df_gals['NED_type'] != '*']
    df_gals = df_gals[df_gals['NED_type'] != '**']
    df_gals = df_gals[df_gals['NED_type'] != 'QSO']
    df_gals = df_gals[df_gals['NED_type'] != 'SNR']
    df_gals = df_gals[df_gals['NED_type'] != 'WD*']
    df_gals = df_gals[df_gals['NED_type'] != '!*']
    df_gals = df_gals[df_gals['NED_type'] != '!V*']
    df_gals = df_gals[df_gals['NED_type'] != 'V*']
    df_gals = df_gals[df_gals['NED_type'] != 'PN']
    df_gals = df_gals[df_gals['NED_type'] != '!Nova']

    #df = df[df['NED_type'] != 'PofG']

    df_gals.reset_index(inplace=True, drop=True)
    df_stars = pd.concat([NED_test_stars, NED_stars]).drop_duplicates()
    #df_stars.to_csv('OSC_061019_PS1_stars_109_NEDCuts.tar.gz')
    #df_gals.to_csv('TNS_PS1_gals_109_NEDCuts.tar.gz')
    return df_gals, df_stars#



# taken from https://github.com/teddyroland/python-biplot/blob/master/biplot.py
import pandas as pd
from sklearn.decomposition import PCA
from matplotlib import pyplot as plt
import os
import numpy as np
import seaborn as sns
from collections import Counter
from matplotlib import cm
import matplotlib.cm as cm
import joypy
from sklearn import preprocessing
import matplotlib
import sys
import joypy

def heatmap(corr, size):
    x = corr['x']
    y = corr['y']

    fig, ax = plt.subplots(figsize=(24,20))

    # Mapping from column names to integer coordinates
    x_labels = [v for v in x.unique()]
    y_labels = [v for v in y.unique()]
    x_to_num = {p[1]:p[0] for p in enumerate(x_labels)}
    y_to_num = {p[1]:p[0] for p in enumerate(y_labels)}

    size_scale = 450
    s = ax.scatter(
        x=x.map(x_to_num), # Use mapping for x
        y=y.map(y_to_num), # Use mapping for y
        s=size * size_scale, # Vector of square sizes, proportional to size parameter
        c=corr['value'],
        cmap='coolwarm',
        marker='s', # Use square as scatterplot marker
        vmin=-1,
        vmax=1
    )
    cbar = fig.colorbar(s, orientation='vertical')
    cbar.ax.tick_params(size=0)
    #cbar.set_label('Correlation', rotation=270)
    cbarax = cbar.ax
    cbarax.text(3,-0.12,'Correlation',rotation=-90, fontsize=30)
    # Show column labels on the axes
    #print([x_to_num[v]+0.5 for v in x_labels])
    plt.xlim((-0.5, 24.5))
    plt.ylim((-0.5, 24.5))
    ax.set_xticks([x_to_num[v]+0.5 for v in x_labels])
    ax.set_xticklabels(x_labels, rotation=90, horizontalalignment='left')
    ax.invert_yaxis()
    ax.set_yticks([y_to_num[v]+0.5 for v in y_labels])
    ##plt.setp( ax.xaxis.get_majorticklabels(), rotation=90, ha="left" )
    ax.set_yticklabels(y_labels)
    dx = -0.3; dy = 0;
    offset_x = matplotlib.transforms.ScaledTranslation(dx, dy, fig.dpi_scale_trans)
    dx = 0; dy = +0.2;
    offset_y = matplotlib.transforms.ScaledTranslation(dx, dy, fig.dpi_scale_trans)

    # apply offset transform to all x ticklabels.
    for label in ax.xaxis.get_majorticklabels():
        label.set_transform(label.get_transform() + offset_x)
    for label in ax.yaxis.get_majorticklabels():
        label.set_transform(label.get_transform() + offset_y)

def preprocess_df(dataML, save=0):
    dataML['rAp-Kron'] = dataML['rApMag_rKronMag']
    del dataML['rApMag_rKronMag']
    #del dataML['7DCD']
    dataML = dataML.drop(['objAltName1','objAltName2','objAltName3'], axis=1)
    dataML = dataML.drop(['objID'], axis=1)
    dataML = dataML.drop(['objName','uniquePspsOBid','ippObjID','surveyID','htmID','zoneID','tessID','projectionID','skyCellID'], axis=1)
    dataML = dataML.drop(['randomID','batchID','dvoRegionID','processingVersion','objInfoFlag','qualityFlag','raStack','decStack'], axis=1)
    dataML = dataML.drop(['raStackErr', 'decStackErr', 'raMean', 'decMean', 'raMeanErr', 'decMeanErr'], axis=1)
    dataML = dataML.drop(['gra', 'gdec', 'graErr', 'gdecErr', 'rra', 'rdec', 'rraErr', 'rdecErr','ira', 'idec', 'iraErr', 'idecErr','zra', 'zdec', 'zraErr', 'zdecErr','yra', 'ydec', 'yraErr', 'ydecErr'], axis=1)
    dataML = dataML.drop(['l','b','nStackObjectRows'],axis=1)
    dataML = dataML.drop(['nStackDetections','nDetections'],axis=1)
    dataML = dataML.drop(['gippDetectID', 'gstackDetectID', 'gstackImageID','rippDetectID', 'rstackDetectID', 'rstackImageID','iippDetectID', 'istackDetectID', 'istackImageID','zippDetectID', 'zstackDetectID', 'zstackImageID','yippDetectID', 'ystackDetectID', 'ystackImageID'], axis=1)
    dataML = dataML.drop(['bestDetection'],axis=1)
    dataML = dataML.drop(['epochMean'],axis=1)
    dataML = dataML.drop(['ng','nr','ni','nz'],axis=1)
    dataML = dataML.drop(['ny'],axis=1)
    dataML = dataML.drop(['uniquePspsSTid','primaryDetection','gEpoch'],axis=1)
    dataML = dataML.drop(['rEpoch','iEpoch','zEpoch', 'yEpoch'],axis=1)
    dataML = dataML.drop(['cx','cy'],axis=1)
    dataML = dataML.drop(['cz'],axis=1)
    #dataML = dataML.drop(['host_logmass', 'host_logmass_min', 'host_logmass_max'],axis=1)
    dataML = dataML.drop(['lambda','beta'],axis=1)
    dataML = dataML.drop(['gpsfChiSq','rpsfChiSq','ipsfChiSq','zpsfChiSq','ypsfChiSq', 'ginfoFlag', 'ginfoFlag2', 'ginfoFlag3',  'rinfoFlag', 'rinfoFlag2', 'rinfoFlag3',  'iinfoFlag', 'iinfoFlag2', 'iinfoFlag3',  'zinfoFlag', 'zinfoFlag2', 'zinfoFlag3',  'yinfoFlag', 'yinfoFlag2', 'yinfoFlag3'],axis=1)
    dataML = dataML.drop(['gxPos', 'gxPosErr','rxPos', 'rxPosErr','ixPos', 'ixPosErr','zxPos', 'zxPosErr','yxPos', 'yxPosErr' ],axis=1)
    dataML = dataML.drop(['gyPos', 'gyPosErr','ryPos', 'ryPosErr','iyPos', 'iyPosErr','zyPos', 'zyPosErr','yyPos', 'yyPosErr' ],axis=1)
    dataML = dataML.drop(['gexpTime','rexpTime','iexpTime','zexpTime','yexpTime','gnFrames','rnFrames','inFrames','znFrames','ynFrames'],axis=1)
    dataML = dataML.drop(['gzp','rzp','izp','zzp','yzp'],axis=1)
    dataML = dataML.drop(['gPlateScale','rPlateScale','iPlateScale','zPlateScale','yPlateScale'],axis=1)
    dataML = dataML.drop(['posMeanChisq'],axis=1)
    dataML = dataML.drop(['gpsfQf','ipsfQf', 'zpsfQf', 'ypsfQf'], axis=1)
    dataML = dataML.drop(['gApFillFac', 'yApFillFac', 'iApFillFac', 'zApFillFac'], axis=1)
    dataML = dataML.drop(['gpsfQfPerfect', 'ipsfQfPerfect', 'zpsfQfPerfect', 'ypsfQfPerfect'], axis=1)
    #dataML = dataML.drop(['level_0'], axis=1)
    dataML = dataML.drop(['gpsfTheta', 'ipsfTheta', 'zpsfTheta', 'ypsfTheta'], axis=1)
    dataML = dataML.drop(['gsky', 'isky', 'zsky', 'ysky'], axis=1)
    dataML = dataML.drop(['gskyErr', 'iskyErr', 'zskyErr', 'yskyErr'], axis=1)
    dataML = dataML.drop(['gpsfCore', 'ipsfCore', 'zpsfCore', 'ypsfCore'], axis=1)
    dataML = dataML.drop(['rpsfTheta', 'rsky', 'rskyErr', 'rpsfCore'], axis=1)
    dataML = dataML.drop(['gpsfLikelihood', 'rpsfLikelihood', 'ipsfLikelihood', 'zpsfLikelihood','ypsfLikelihood'], axis=1)
    dataML = dataML.drop(['rpsfQf'], axis=1)
    dataML = dataML.drop(['rpsfQfPerfect'], axis=1)
    dataML = dataML.drop(['rApFillFac'], axis=1)
    #dataML.drop(['objID'], inplace=True, axis=1)
    dataML = dataML.drop(['NED_redshift', 'NED_type', 'NED_mag', 'NED_name', 'NED_vel', 'TransientDEC', 'TransientDiscoveryDate', 'TransientDiscoveryMag', 'TransientDiscoveryYear', 'TransientRA'], axis=1)
    dataML = dataML.drop(['TransientRedshift',
       'TransientRedshift', 'Transient AltName',
       'host_logmass', 'host_logmass_min', 'host_logmass_max',
       'Hubble Residual'], axis=1)

    dataML = dataML.dropna()

    # Labels are the values we want to predict
    dataML.loc[dataML['TransientClass'] == 'SN Ib\n SN Ib', 'TransientClass'] = 'SN Ib'
    dataML.loc[dataML['TransientClass'] == 'SN Ia\n SN Ia', 'TransientClass'] = 'SN Ia'
    dataML.loc[dataML['TransientClass'] == 'SN II-pec', 'TransientClass'] = 'SN II Pec'
    dataML.loc[dataML['TransientClass'] == 'SN Ic-BL', 'TransientClass'] = 'SN Ic'
    dataML.loc[dataML['TransientClass'] == 'SN Ia-91T-like', 'TransientClass'] = 'SN Ia'
    dataML.loc[dataML['TransientClass'] == 'SN Ia-pec', 'TransientClass'] = 'SN Ia Pec'
    dataML.loc[dataML['TransientClass'] == 'SN Ib-pec', 'TransientClass'] = 'SN Ib'
    dataML.loc[dataML['TransientClass'] == 'SN Ic', 'TransientClass'] = 'SN Ib/c'
    dataML.loc[dataML['TransientClass'] == 'SN Iax[02cx-like]', 'TransientClass'] = 'SN Ia'
    dataML.loc[dataML['TransientClass'] == 'SN Ib-Ca-rich', 'TransientClass'] = 'SN Ib'
    dataML.loc[dataML['TransientClass'] == 'SN Ia-91bg-like', 'TransientClass'] = 'SN Ia'
    dataML.loc[dataML['TransientClass'] == 'SN Ia-CSM', 'TransientClass'] = 'SN Ia'
    dataML.loc[dataML['TransientClass'] == 'SN Ic-pec', 'TransientClass'] = 'SN Ic'
    dataML.loc[dataML['TransientClass'] == 'SN IIn', 'TransientClass'] = 'SN II'
    dataML.loc[dataML['TransientClass'] == 'SN Ibn', 'TransientClass'] = 'SN Ib'
    dataML.loc[dataML['TransientClass'] == 'SN Ib', 'TransientClass'] = 'SN Ib/c'
    dataML.loc[dataML['TransientClass'] == 'SN Ia Pec', 'TransientClass'] = 'SN Ia'
    dataML.loc[dataML['TransientClass'] == 'SN Ia-91bg', 'TransientClass'] = 'SN Ia'
    dataML.loc[dataML['TransientClass'] == 'SN Ic', 'TransientClass'] = 'SN Ib/c'
    #dataML.loc[dataML['TransientClass'] == 'SN Ib/c', 'TransientClass'] = 'SN Ib'
    dataML.loc[dataML['TransientClass'] == 'SN IIP', 'TransientClass'] = 'SN II'
    dataML.loc[dataML['TransientClass'] == 'SN IIL', 'TransientClass'] = 'SN II'
    #dataML.loc[dataML['TransientClass'] == 'SLSN-I', 'TransientClass'] = 'SLSN'
    #dataML.loc[dataML['TransientClass'] == 'SN IIb', 'TransientClass'] = 'SN II'
    dataML.loc[dataML['TransientClass'] == 'SN I', 'TransientClass'] = 'SN I?'
    #dataML.loc[dataML['TransientClass'] == 'SN II', 'TransientClass'] = 'Core Collapse'
    dataML.loc[dataML['TransientClass'] == 'SN Ib', 'TransientClass'] = 'SN Ib/c'
    dataML.loc[dataML['TransientClass'] == 'SN Ic', 'TransientClass'] = 'SN Ib/c'
    #dataML.loc[dataML['TransientClass'] == 'SN Ib/c', 'TransientClass'] = 'Core Collapse'
    dataML.loc[dataML['TransientClass'] == 'SLSN-I', 'TransientClass'] = 'SLSN'
    dataML.loc[dataML['TransientClass'] == 'SLSN-II', 'TransientClass'] = 'SLSN'
    dataML.loc[dataML['TransientClass'] == 'Ia Pec', 'TransientClass'] = 'SN Ia Pec'
    dataML.loc[dataML['TransientClass'] == 'II Pec', 'TransientClass'] = 'SN II Pec'
    dataML.loc[dataML['TransientClass'] == 'Ia*', 'TransientClass'] = 'SN Ia'
    dataML.loc[dataML['TransientClass'] == 'Ia-02cx', 'TransientClass'] = 'SN Ia'
    dataML.loc[dataML['TransientClass'] == 'Ia-91T', 'TransientClass'] = 'SN Ia-91T-like'
    dataML.loc[dataML['TransientClass'] == 'Ia-91bg', 'TransientClass'] = 'SN Ia'
    dataML.loc[dataML['TransientClass'] == 'Ia-99aa', 'TransientClass'] = 'SN Ia'
    dataML.loc[dataML['TransientClass'] == 'SLSN-II', 'TransientClass'] = 'SLSN'
    dataML.loc[dataML['TransientClass'] == 'CC', 'TransientClass'] = 'SN II'
    dataML.loc[dataML['TransientClass'] == 'II', 'TransientClass'] = 'SN II'
    dataML.loc[dataML['TransientClass'] == 'SLSN-I-R', 'TransientClass'] = 'SLSN'
    dataML.loc[dataML['TransientClass'] == 'SLSN-R', 'TransientClass'] = 'SLSN'
    dataML.loc[dataML['TransientClass'] == 'II/IIb', 'TransientClass'] = 'SN II'
    #dataML.loc[dataML['TransientClass'] == 'II L', 'TransientClass'] = 'SN II'
    dataML.loc[dataML['TransientClass'] == 'II P', 'TransientClass'] = 'SN IIP'
    dataML.loc[dataML['TransientClass'] == 'Ib', 'TransientClass'] = 'SN Ib/c'
    dataML.loc[dataML['TransientClass'] == 'Ic', 'TransientClass'] = 'SN Ib/c'
    dataML.loc[dataML['TransientClass'] == 'II-p', 'TransientClass'] = 'SN II P'
    #dataML.loc[dataML['TransientClass'] == 'II/LBV', 'TransientClass'] = 'SN II'
    dataML.loc[dataML['TransientClass'] == 'IIb', 'TransientClass'] = 'SN IIb'
    #dataML.loc[dataML['TransientClass'] == 'Ic Pec', 'TransientClass'] = 'SN Ib/c'
    dataML.loc[dataML['TransientClass'] == 'SN Ia Pec', 'TransientClass'] = 'SN Ia'
    #dataML.loc[dataML['TransientClass'] == 'Ib/Ic (Ca rich?)?', 'TransientClass'] = 'SN Ib/c'
    #dataML.loc[dataML['TransientClass'] == 'Ia CSM', 'TransientClass'] = 'SN Ia'
    #dataML.loc[dataML['TransientClass'] == 'Ic BL', 'TransientClass'] = 'SN Ib/c'
    dataML.loc[dataML['TransientClass'] == 'Ia', 'TransientClass'] = 'SN Ia'
    dataML.loc[dataML['TransientClass'] == 'Ib/c', 'TransientClass'] = 'SN Ib/c'
    dataML.loc[dataML['TransientClass'] == 'IIn', 'TransientClass'] = 'SN IIn'
    #dataML.loc[dataML['TransientClass'] == 'Ib Pec', 'TransientClass'] = 'SN Ib/c'
    #dataML.loc[dataML['TransientClass'] == 'Ibn', 'TransientClass'] = 'SN Ib/c'
    #dataML.loc[dataML['TransientClass'] == 'IIn Pec', 'TransientClass'] = 'SN IIn'
    dataML.loc[dataML['TransientClass'] == 'Ia/Ic', 'TransientClass'] = 'SN Ia/c'
    dataML.loc[dataML['TransientClass'] == 'SN II P', 'TransientClass'] = 'SN IIP'
    dataML.loc[dataML['TransientClass'] == 'SN II Pec', 'TransientClass'] = 'SN II'
    dataML.loc[dataML['TransientClass'] == 'Ia/c', 'TransientClass'] = 'SN Ia/c'
    dataML.loc[dataML['TransientClass'] == 'I', 'TransientClass'] = 'SN I'
    #dataML.loc[dataML['TransientClass'] == 'SN Ia-91T-like', 'TransientClass'] = 'SN Ia Pec'
    #dataML.loc[dataML['TransientClass'] == 'SN Iax[02cx-like]', 'TransientClass'] = 'SN Ia Pec'

    #dataML.loc[dataML['TransientClass'] == 'SLSN-I?', 'TransientClass'] = 'SLSN'

    #put that back in there, along with the SLSN-I-R
    #dataML.loc[dataML['TransientClass'] == 'SLSN-IIn', 'TransientClass'] = 'SLSN'

    # only for the 2 component class
    #dataML.loc[dataML['TransientClass'] == 'SN Ib/c', 'TransientClass'] = 'Core Collapse'
    #dataML.loc[dataML['TransientClass'] == 'SN IIP', 'TransientClass'] = 'Core Collapse'
    #dataML.loc[dataML['TransientClass'] == 'SN IIb', 'TransientClass'] = 'Core Collapse'
    #dataML.loc[dataML['TransientClass'] == 'SN IIn', 'TransientClass'] = 'Core Collapse'
    #dataML.loc[dataML['TransientClass'] == 'SN II', 'TransientClass'] = 'Core Collapse'

    #'SN Ia': 6279, 'SN II': 2061, 'SN Ib/c': 528, 'SN IIP': 307, 'SN IIn': 265, 'SN IIb': 94, 'SLSN': 38

    # 'SN II', 'SN IIb', 'SN IIn',
    # Delete the rows where we have no set detection
    dataML = dataML[dataML['TransientClass'] != 'SN']
    #dataML = dataML[dataML['TransientClass'] != 'SLSN']
    dataML = dataML[dataML['TransientClass'] != 'Ia-09dc']
    dataML = dataML[dataML['TransientClass'] != 'SN I?']
    dataML = dataML[dataML['TransientClass'] != 'SNIa?']
    dataML = dataML[dataML['TransientClass'] != 'SLSN?']
    dataML = dataML[dataML['TransientClass'] != 'Ic Pec']
    dataML = dataML[dataML['TransientClass'] != 'PISN?']
    dataML = dataML[dataML['TransientClass'] != 'I?']
    dataML = dataML[dataML['TransientClass'] != 'Ib/Ic (Ca rich?)?']
    dataML = dataML[dataML['TransientClass'] != 'II Pec?']
    dataML = dataML[dataML['TransientClass'] != 'IIn?']
    dataML = dataML[dataML['TransientClass'] != 'IIb?']
    dataML = dataML[dataML['TransientClass'] != 'SLSN?']
    dataML = dataML[dataML['TransientClass'] != 'SN Ic-BL']

    #dataML = dataML[dataML['TransientClass'] != 'SN IIb']
    #dataML = dataML[dataML['TransientClass'] != 'SN Ia Pec']
    #dataML = dataML[dataML['TransientClass'] != 'SN IIn']

    #'SLSN', 'SN II', 'SN II Pec', 'SN IIP', 'SN IIb', 'SN IIn',
    #       'SN Ia', 'SN Ia Pec', 'SN Ia-91T-like', 'SN Ib/c'

    dataML = dataML[dataML['TransientClass'] != 'Ic?']
    dataML = dataML[dataML['TransientClass'] != 'II?']
    dataML = dataML[dataML['TransientClass'] != 'Ib/IIb']
    dataML = dataML[dataML['TransientClass'] != 'IIb/Ib']
    dataML = dataML[dataML['TransientClass'] != 'II/Ib/c']
    dataML = dataML[dataML['TransientClass'] != 'Ib/c?']
    dataML = dataML[dataML['TransientClass'] != 'SLSN-II?']
    dataML = dataML[dataML['TransientClass'] != 'Ia?']
    # only for the 4-component class
    dataML = dataML[dataML['TransientClass'] != 'SN I']

    dataML = dataML[dataML['TransientClass'] != 'LBV to IIn']
    dataML = dataML[dataML['TransientClass'] != 'LBV']
    dataML = dataML[dataML['TransientClass'] != 'SN Ia/c']

    dataML = dataML[dataML['TransientClass'] != 'Ca-rich']
    dataML = dataML[dataML['TransientClass'] != 'Pec']
    dataML = dataML[dataML['TransientClass'] != 'CN']
    dataML = dataML[dataML['TransientClass'] != 'II L?']
    dataML = dataML[dataML['TransientClass'] != 'Ib-Ca']
    dataML = dataML[dataML['TransientClass'] != 'Pec']
    dataML = dataML[dataML['TransientClass'] != 'nIa']
    dataML = dataML[dataML['TransientClass'] != 'SLSN-I?']
    dataML = dataML[dataML['TransientClass'] != 'SLSN-IIn']
    dataML = dataML[dataML['TransientClass'] != 'SN Iax[02cx-like]']
    dataML = dataML[dataML['TransientClass'] != 'SN Ia-91T-like']
    dataML = dataML[dataML['TransientClass'] != 'Ia-02cx']
    dataML = dataML[dataML['TransientClass'] != 'Ia-91T']
    dataML = dataML[dataML['TransientClass'] != 'Ia-91bg']
    dataML = dataML[dataML['TransientClass'] != 'IIn Pec']

    dataML_orig = dataML

    dataML = dataML.drop(['TransientName'],axis=1)
    dataML = dataML.drop(['TransientClass'],axis=1)

    dataML.dropna(inplace=True)

    if save:
        dataML.to_csv("pre_PCA_features.csv",index=False)
    return dataML_orig, dataML

def plot_heatmap(dataML, save=0, corr_type='pearson'):
    dataML["gSNR"] = 1/dataML["gApMagErr"]
    dataML["rSNR"] = 1/dataML["rApMagErr"]
    dataML["iSNR"] = 1/dataML["iApMagErr"]
    dataML["zSNR"] = 1/dataML["zApMagErr"]
    dataML["ySNR"] = 1/dataML["yApMagErr"]

    dataML['4DCD'] = dataML['7DCD']
    dataML[r'$\theta$'] = dataML['dist']
    dataML[r'$\theta/d_{DLR}$'] = dataML['dist/DLR']
    dataML[r'Ap - Kron'] = dataML['gApMag_gKronMag']
    dataML[r'KronRad'] = dataML['gKronRad']
    dataML[r'PSFMag'] = dataML['gPSFMag']
    dataML[r'PSFMagErr'] = dataML['gPSFMagErr']
    dataML[r'ApMag'] = dataML['gApMag']
    dataML[r'ApMagErr'] = dataML['gApMagErr']
    dataML[r'KronMag'] = dataML['gKronMag']
    dataML[r'KronMagErr'] = dataML['gKronMagErr']
    dataML[r'psfMajorFWHM'] = dataML['gpsfMajorFWHM']
    dataML[r'psfMinorFWHM'] = dataML['gpsfMinorFWHM']
    dataML[r'momentXX'] = dataML['gmomentXX']
    dataML[r'momentXY'] = dataML['gmomentXY']
    dataML[r'momentYY'] = dataML['gmomentYY']

    dataML[r'momentR1'] = dataML['gmomentR1']
    dataML[r'momentRH'] = dataML['gmomentRH']
    dataML[r'ApRadius'] = dataML['gApRadius']
    dataML[r'ExtNSigma'] = dataML['gExtNSigma']
    dataML[r'PSFFlux'] = dataML['gPSFFlux']
    dataML[r'PSFFluxErr'] = dataML['gPSFFluxErr']
    dataML[r'ApFlux'] = dataML['gApFlux']
    dataML[r'ApFluxErr'] = dataML['gApFluxErr']
    dataML[r'KronFlux'] = dataML['gKronFlux']
    dataML[r'KronFluxErr'] = dataML['gKronFluxErr']

    dataML_corr = dataML[[r'PSFMag',r'PSFMagErr','ApMag','ApMagErr',r'KronMag','KronMagErr','psfMajorFWHM','psfMinorFWHM',
    'momentXX','momentXY','momentYY','momentR1','momentRH','PSFFlux','PSFFluxErr','ApFlux','ApFluxErr','ApRadius',
    'KronFlux','KronFluxErr',r'KronRad','ExtNSigma',r'Ap - Kron',
    'g-r','r-i','i-z','z-y','4DCD',r'$\theta$',r'$\theta/d_{DLR}$',r'g-rErr']]

    dataML.drop(['rAp-Kron', 'ApRadius', 'Ap - Kron','KronRad','PSFMag','PSFMagErr','ApMag','ApMagErr','KronMag','KronMagErr','psfMajorFWHM','psfMinorFWHM','momentXX','momentXY','momentYY','momentR1','momentRH'],axis=1, inplace=True)
    dataML.drop(['ExtNSigma','PSFFlux','PSFFluxErr','ApFlux','ApFluxErr','KronFlux','KronFluxErr', '4DCD', r'$\theta$', r'$\theta/d_{DLR}$'], axis=1, inplace=True)

    cols = dataML.columns.values
    cols = ['g-r', 'gApFlux', 'gApFluxErr', 'gApMag', 'gApMagErr', #g
       'gApMag_gKronMag', 'gApRadius', 'gExtNSigma', 'gKronFlux',
       'gKronFluxErr', 'gKronMag', 'gKronMagErr', 'gKronRad', 'gPSFFlux',
       'gPSFFluxErr', 'gPSFMag', 'gPSFMagErr', 'gmomentR1', 'gmomentRH',
       'gmomentXX', 'gmomentXY', 'gmomentYY', 'gpsfMajorFWHM',
       'gpsfMinorFWHM','g-rErr',
       'r-i', 'rApFlux', 'rApFluxErr', #r
       'rApMag', 'rApMagErr', 'rApRadius', 'rExtNSigma', 'rKronFlux',
       'rKronFluxErr', 'rKronMag', 'rKronMagErr', 'rKronRad', 'rPSFFlux',
       'rPSFFluxErr', 'rPSFMag', 'rPSFMagErr', 'rmomentR1', 'rmomentRH',
       'rmomentXX', 'rmomentXY', 'rmomentYY', 'rpsfMajorFWHM',
       'rpsfMinorFWHM','r-iErr',
       'i-z', 'iApFlux', 'iApFluxErr', 'iApMag', #i
       'iApMagErr', 'iApMag_iKronMag', 'iApRadius', 'iExtNSigma',
       'iKronFlux', 'iKronFluxErr', 'iKronMag', 'iKronMagErr', 'iKronRad',
       'iPSFFlux', 'iPSFFluxErr', 'iPSFMag', 'iPSFMagErr', 'imomentR1',
       'imomentRH', 'imomentXX', 'imomentXY', 'imomentYY',
       'ipsfMajorFWHM', 'ipsfMinorFWHM','i-zErr',
       'z-y','zApFlux', 'zApFluxErr', 'zApMag', 'zApMagErr', #z
       'zApMag_zKronMag', 'zApRadius', 'zExtNSigma', 'zKronFlux',
       'zKronFluxErr', 'zKronMag', 'zKronMagErr', 'zKronRad', 'zPSFFlux',
       'zPSFFluxErr', 'zPSFMag', 'zPSFMagErr', 'zmomentR1', 'zmomentRH',
       'zmomentXX', 'zmomentXY', 'zmomentYY', 'zpsfMajorFWHM',
       'zpsfMinorFWHM','z-yErr',
       'yApFlux', 'yApFluxErr', 'yApMag', 'yApMagErr', #y
       'yApMag_yKronMag', 'yApRadius', 'yExtNSigma', 'yKronFlux',
       'yKronFluxErr', 'yKronMag', 'yKronMagErr', 'yKronRad', 'yPSFFlux',
       'yPSFFluxErr', 'yPSFMag', 'yPSFMagErr', 'ymomentR1', 'ymomentRH',
       'ymomentXX', 'ymomentXY', 'ymomentYY', 'ypsfMajorFWHM',
       'ypsfMinorFWHM', 'dist/DLR', 'dist']

    dataML_shifted = dataML[cols]


    sns.set_context("poster")
    plt.figure(figsize=(250, 200))
    sns.heatmap(dataML_shifted.corr(method=corr_type), annot=False, cmap='coolwarm', vmin=-1, vmax=1, linecolor='white', linewidths=0.3,annot_kws={"fontsize": "5"},cbar_kws={'label': 'Correlation'})
    if save:
        plt.savefig("heatmap_fullTable_%s.png"%corr_type, bbox_inches='tight')
        dataML.corr(method=corr_type).to_csv("GHOST_fullTable_correlations_%s.tar.gz"%corr_type, index=False)

    #dataML_corr_sorted = dataML_corr[['g-r', 'momentR1', 'KronRad', 'Ap - Kron', 'momentXX', '4DCD',
    #    'ExtNSigma', 'PSFFlux', 'r-i', 'momentYY','momentRH',
    #    'PSFMag', 'PSFMagErr', 'ApMag',
    #    'ApMagErr', 'KronMag',
    #    'KronMagErr', 'psfMajorFWHM', 'psfMinorFWHM',
    #    'momentXY',
    #    'PSFFluxErr', 'ApFlux', 'ApFluxErr', 'ApRadius', 'KronFlux',
    #    'KronFluxErr',
    #    'i-z', 'z-y', '$\\theta$', '$\\theta/d_{DLR}$']]
    dataML_corr_sorted = dataML_corr[['g-r', 'ApFlux', 'ApFluxErr', 'ApMag', 'ApMagErr', #g
       'Ap - Kron', 'ApRadius', 'ExtNSigma', 'KronFlux',
       'KronFluxErr', 'KronMag', 'KronMagErr', 'KronRad', 'PSFFlux',
       'PSFFluxErr', 'PSFMag', 'PSFMagErr', 'momentR1', 'momentRH',
       'momentXX', 'momentXY', 'momentYY', 'psfMajorFWHM',
       'psfMinorFWHM','g-rErr']]

    data = dataML_corr_sorted

    columns = dataML_corr_sorted.columns
    corr = data[columns].corr(method=corr_type)
    corr = pd.melt(corr.reset_index(), id_vars='index') # Unpivot the dataframe, so we can get pair of arrays for x and y
    corr.columns = ['x', 'y', 'value']
    heatmap(
        corr,
        size=corr['value'].abs()
    )
    if save:
        plt.savefig("heatmap_withScaledSquares_%s.png" %corr_type, dpi=300, bbox_inches='tight')

    #matrix = np.triu(dataML_corr.dropna().corr(method=corr_type))
    #plt.figure(figsize=(20, 14))
    #sns.heatmap(dataML_corr.corr(method=corr_type), annot=False, mask=matrix, cmap='coolwarm', vmin=-1, vmax=1, linecolor='white', linewidths=0.3,annot_kws={"fontsize": "30"},cbar_kws={'label': 'Correlation'})
    #if save:
    #    plt.savefig("heatmap_triangle_%s.pdf"%corr_type,dpi=300, bbox_inches='tight')

    dataML = dataML.drop(['gSNR', 'rSNR', 'iSNR', 'zSNR', 'ySNR'], axis=1)
    dataML_sub = dataML[['r-i', 'iApMag_iKronMag', 'zApMag_zKronMag', 'g-r']]
    bestFeatures = np.array(['r-i', 'iApMag_iKronMag', 'zApMag_zKronMag', 'g-r', 'gApMag_gKronMag', 'yApMag_yKronMag', 'gExtNSigma', '7DCD', 'rPSFMag_rKronMag', 'yExtNSigma', 'zExtNSigma', 'iPSFMag_zPSFMag', 'gmomentRH', 'i-z'])
    #del dataML['Unnamed: 0']
    dataML['4DCD'] = dataML['7DCD']
    del dataML['7DCD']
    #get rid of color for now
    del dataML['4DCD']
    dataML[r'$\theta$'] = dataML['dist']
    del dataML['dist']
    dataML[r'$\theta/d_{DLR}$'] = dataML['dist/DLR']
    del dataML['dist/DLR']

def tsne_ghost(dataML):
    dataML_orig, dataML = preprocess_df(dataML)
    plot_heatmap(dataML, save=0, corr_type='spearman')

    #scaling with standardscaler
    dataML_scaled = preprocessing.scale(dataML)
    ## perform PCA
    n = len(dataML.columns)
    pca = PCA(n_components = 2)
    df_plot = pca.fit(dataML_scaled)

    ## project data into PC space
    xvector = pca.components_[0] # see 'prcomp(my_data)$rotation' in R
    yvector = pca.components_[1]

    xs = pca.transform(dataML_scaled)[:,0] # see 'prcomp(my_data)$x' in R
    ys = pca.transform(dataML_scaled)[:,1]

    plt.figure(figsize=(10,7))
    sns.kdeplot(df_plot.loc[df_plot['TransientClass']=='SN Ia','xs'], shade=True, shade_lowest=False, alpha=0.6, label='SN Ia',color='tab:blue')
    plt.axvline(np.median(df_plot.loc[df_plot['TransientClass']=='SN Ia','xs']), linestyle='--',color='tab:blue')
    sns.kdeplot(df_plot.loc[df_plot['TransientClass']=='SN Ib/c','xs'], shade=True, shade_lowest=False, alpha=0.6, label='SN Ib/c',color='tab:green')
    plt.axvline(np.median(df_plot.loc[df_plot['TransientClass']=='SN Ib/c','xs']), linestyle='--',color='tab:green')
    sns.kdeplot(df_plot.loc[df_plot['TransientClass']=='SLSN','xs'], shade=True, shade_lowest=False, alpha=0.6, label='SLSN', color='tab:orange')
    plt.axvline(np.median(df_plot.loc[df_plot['TransientClass']=='SLSN','xs']), linestyle='--',color='tab:orange')
    plt.legend(fontsize=16)
    plt.xlabel("PC1 (71.5%)",fontsize=16)
    plt.xlim((-20,20))
    plt.savefig("PCA_axis1Only_withMedians_scaler.png", dpi=300)

    Counter(df_plot['TransientClass'])
    dropClass = np.array(['II L', 'II/LBV', 'Ia CSM', 'Ib Pec', 'Ibn', 'Ic BL','SN IIL','SN Ia-91bg-like', 'SN Ia-CSM', 'SN Ib-Ca-rich', 'SN Ib-pec', 'SN Ibn', 'SN Ic-pec'])
    df_plot = df_plot[~df_plot['TransientClass'].isin(dropClass)]

    # red purple blue
    Counter(df_plot['TransientClass'])
    cols = ['#ff6666', '#861388', '#0d3b66']
    #classes = ['SN Ia', 'SN II','SN Ib/c']
    #classes = ['SN Ib/c','SN II','SN Ia']
    classes = ['SN Ib/c','SLSN', 'SN Ia']

    length = np.sqrt(xvector**2 + yvector**2)

    def checkFeature(feature):
        if feature.endswith("Err"):
            return False
        if feature.startswith("4"):
            return False
        if (not feature.startswith('g') and  not feature.startswith('i') and not feature.startswith('z') and not feature.startswith('y')):
            return True
        return False

    sns.set_context("poster")
    #pl = sns.color_palette(['#f77189', '#3ba3ec''#50b131'])
    pl = [sns.light_palette("firebrick", as_cmap=True), sns.light_palette("purple", as_cmap=True), sns.light_palette("orange", as_cmap=True)]
    plt.figure(figsize=(10,10))
    plt.xlim((-75, 75))
    plt.ylim((-30,40))
    for i in np.arange(len(cols)):
        if i == 2:
            shadeval = False
            ls = '--'
        else:
            shadeval=True
            ls = '-'
        alpha=1.
        sns.kdeplot(df_plot.loc[df_plot['TransientClass'] == classes[i], 'xs'], df_plot.loc[df_plot['TransientClass'] == classes[i],'ys'], alpha=alpha, linewidths=5, n_levels=5, linestyles=ls,shade=shadeval, shade_lowest=False,label=classes[i])
    facs = ['rKronMag', 'rKronMag', 'rmomentR1', 'rExtNSigma']
    j = 0

    for i in range(len(xvector)):
    # arrows project features (ie columns from csv) as vectors onto PC axes
        length = np.sqrt(xvector[i]**2 + yvector[i]**2)
        if length > 0.1:
            if checkFeature(dataML.columns.values[i]):
                if dataML.columns.values[i].endswith("Mag"):
                    xtext = xvector[i]*max(xs) - 9
                    ytext = yvector[i]*max(ys) - 0.25
                elif dataML.columns.values[i].endswith("Flux"):
                    xtext = xvector[i]*max(xs) + 0.5
                    ytext = yvector[i]*max(ys)
                elif dataML.columns.values[i] == 'rmomentR1':
                    xtext = xvector[i]*max(xs)+1.5
                    ytext = yvector[i]*max(ys) - 1.8
                elif dataML.columns.values[i] == 'rExtNSigma':
                    xtext = xvector[i]*max(xs)
                    ytext = yvector[i]*max(ys) - 2
                elif dataML.columns.values[i] == 'rmomentRH':
                    xtext = xvector[i]*max(xs) + 0.5
                    ytext = yvector[i]*max(ys) - 0.75
                elif dataML.columns.values[i] == 'rKronRad':
                    xtext = xvector[i]*max(xs)+2.5
                    ytext = yvector[i]*max(ys)
                elif dataML.columns.values[i] == 'rAp-Kron':
                    #dataML.columns.values[i] = 'rAp-Kron'
                    xtext = xvector[i]*max(xs) + 2
                    ytext = yvector[i]*max(ys) - 0.5
                elif dataML.columns.values[i] == 'rmomentXX':
                    xtext = xvector[i]*max(xs) + 2
                    ytext = yvector[i]*max(ys) - 1.5
                elif dataML.columns.values[i] == 'rmomentYY':
                    xtext = xvector[i]*max(xs) + 2
                    ytext = yvector[i]*max(ys) - 0.75
                else:
                    xtext = xvector[i]*max(xs)
                    ytext = yvector[i]*max(ys)

                if dataML.columns.values[i] == 'rKronFlux':
                    xtext = xvector[i]*max(xs) + 0.5
                    ytext = yvector[i]*max(ys) -1
                elif dataML.columns.values[i] == 'rKronMag':
                    xtext = xvector[i]*max(xs) - 13
                    ytext = yvector[i]*max(ys) - 0.25
                plt.text(xtext, ytext,
                    list(dataML.columns.values)[i], fontsize=26, color='#2e2e3a')
        #    j += 1
                plt.arrow(0, 0, xvector[i]*max(xs), yvector[i]*max(ys),
                color='#2e2e3a', width=0.05, head_width=0.25, zorder=10, alpha=1, lw=3)
    #plt.text(-5, -22,'SN Ia', fontsize=20, color='#ff6666')
    #plt.text(-60, 10,'SN II', fontsize=20, color='#861388')
    #plt.text(55, 20,'SN Ib/c', fontsize=20, color='#0d3b66')
    #re-calculate DLR!
    plt.xlim((-22, 20))
    plt.ylim((-12, 12))
    plt.xlabel("PC 1 (44.5%)",fontsize=20)
    #0.44575393, 0.10369499
    plt.ylabel("PC 2 (10.4%)",fontsize=20)
    plt.legend(fontsize=20, loc='lower left')
    plt.savefig("PCA_Ia_Ibc_II_scaler.png", dpi=300,bbox_inches='tight')

    fig = plt.figure(figsize = (8,8))
    plt.plot(np.arange(10), pca.explained_variance_ratio_,'o--',markersize=15)
    plt.xticks(np.arange(10))
    plt.xlabel("Principal Component",fontsize=18)
    plt.ylabel("Explained Variance (%)",fontsize=18)
    np.sum(pca.explained_variance_ratio_)

    #create data frame with xs, ys, dataML_orig['TransientClass']
    dataML_orig.reset_index(drop=True, inplace=True)

    df_plot_subset = df_plot[['ys', 'TransientClass']]
    df_plot_subset2 = df_plot[['xs', 'TransientClass']]

    sns.set_context("notebook", font_scale=1.5)
    fig, axes = joypy.joyplot(df_plot_subset2, by="TransientClass", ylim='own',fade=True, colormap=cm.Dark2, kind="kde")
    fig.set_figheight(10)
    fig.set_figwidth(10)
    axes[0].set_xlim((-25,25))
    fig.savefig("PCA2_joyplot2_scaler_dropOutliers.png", dpi=200)

    df_plot_lowZ = df_plot.loc[df_plot['TransientRedshift'] < 0.1]
    df_plot_subset_lowZ = df_plot_lowZ[['ys', 'TransientClass']]
    df_plot_subset_lowZ2 = df_plot_lowZ[['xs', 'TransientClass']]

    fig, axes = joypy.joyplot(df_plot_subset_lowZ2, by="TransientClass", ylim='own', fade=True, colormap=cm.Dark2, kind="kde")
    fig.set_figheight(10)
    fig.set_figwidth(10)
    axes[0].set_xlim(xmin=-20, xmax=20)
    fig.savefig("PCA2_joyplot2_lowZ_dropOutliers.pdf")

    fig = plt.figure(figsize = (8,8))
    plt.plot(np.arange(2), pca.explained_variance_ratio_,'o--',markersize=15)
    plt.xticks(np.arange(2))
    plt.xlabel("Principal Component",fontsize=18)
    plt.ylabel("Explained Variance (%)",fontsize=18)
    plt.savefig("Explained_variance.png", dpi=300, bbox_inches='tight')



import numpy as np
from astropy.table import Table
import requests
from PIL import Image
from io import BytesIO
import pathlib
import os
import sys
import pandas as pd
import matplotlib.pyplot as plt
from astropy.io import fits
from astropy import units as u
from astropy.coordinates import SkyCoord
import re
from astro_ghost.PS1QueryFunctions import getcolorim, get_hosts
from astro_ghost.hostMatching import build_ML_df
from astro_ghost.NEDQueryFunctions import getNEDInfo
from astro_ghost.gradientAscent import gradientAscent
from astro_ghost.starSeparation import separateStars_STRM
from astro_ghost.sourceCleaning import clean_dict, removePS1Duplicates, getColors, makeCuts
from astro_ghost.stellarLocus import calc_7DCD
from astro_ghost.DLR import chooseByDLR
import requests
import pickle
import os
import glob
from datetime import datetime

def getGHOST(real=False, verbose=False):
    if not os.path.exists('./database'):
        os.makedirs('./database')
    if real:
        url = 'http://ghost.ncsa.illinois.edu/static/database/GHOST.csv'
        r = requests.get(url)
        fname = './database/GHOST.csv'
        open(fname , 'wb').write(r.content)
        if verbose:
            print("Successfully downloaded GHOST database from ghost.ncsa.illinois.edu.\n")
    else:
        #create dummy database
        colnames = ['objName', 'objAltName1', 'objAltName2', 'objAltName3', 'objID',
       'uniquePspsOBid', 'ippObjID', 'surveyID', 'htmID', 'zoneID',
       'tessID', 'projectionID', 'skyCellID', 'randomID', 'batchID',
       'dvoRegionID', 'processingVersion', 'objInfoFlag', 'qualityFlag',
       'raStack', 'decStack', 'raStackErr', 'decStackErr', 'raMean',
       'decMean', 'raMeanErr', 'decMeanErr', 'epochMean', 'posMeanChisq',
       'cx', 'cy', 'cz', 'lambda', 'beta', 'l', 'b', 'nStackObjectRows',
       'nStackDetections', 'nDetections', 'ng', 'nr', 'ni', 'nz', 'ny',
       'uniquePspsSTid', 'primaryDetection', 'bestDetection',
       'gippDetectID', 'gstackDetectID', 'gstackImageID', 'gra', 'gdec',
       'graErr', 'gdecErr', 'gEpoch', 'gPSFMag', 'gPSFMagErr', 'gApMag',
       'gApMagErr', 'gKronMag', 'gKronMagErr', 'ginfoFlag', 'ginfoFlag2',
       'ginfoFlag3', 'gnFrames', 'gxPos', 'gyPos', 'gxPosErr', 'gyPosErr',
       'gpsfMajorFWHM', 'gpsfMinorFWHM', 'gpsfTheta', 'gpsfCore',
       'gpsfLikelihood', 'gpsfQf', 'gpsfQfPerfect', 'gpsfChiSq',
       'gmomentXX', 'gmomentXY', 'gmomentYY', 'gmomentR1', 'gmomentRH',
       'gPSFFlux', 'gPSFFluxErr', 'gApFlux', 'gApFluxErr', 'gApFillFac',
       'gApRadius', 'gKronFlux', 'gKronFluxErr', 'gKronRad', 'gexpTime',
       'gExtNSigma', 'gsky', 'gskyErr', 'gzp', 'gPlateScale',
       'rippDetectID', 'rstackDetectID', 'rstackImageID', 'rra', 'rdec',
       'rraErr', 'rdecErr', 'rEpoch', 'rPSFMag', 'rPSFMagErr', 'rApMag',
       'rApMagErr', 'rKronMag', 'rKronMagErr', 'rinfoFlag', 'rinfoFlag2',
       'rinfoFlag3', 'rnFrames', 'rxPos', 'ryPos', 'rxPosErr', 'ryPosErr',
       'rpsfMajorFWHM', 'rpsfMinorFWHM', 'rpsfTheta', 'rpsfCore',
       'rpsfLikelihood', 'rpsfQf', 'rpsfQfPerfect', 'rpsfChiSq',
       'rmomentXX', 'rmomentXY', 'rmomentYY', 'rmomentR1', 'rmomentRH',
       'rPSFFlux', 'rPSFFluxErr', 'rApFlux', 'rApFluxErr', 'rApFillFac',
       'rApRadius', 'rKronFlux', 'rKronFluxErr', 'rKronRad', 'rexpTime',
       'rExtNSigma', 'rsky', 'rskyErr', 'rzp', 'rPlateScale',
       'iippDetectID', 'istackDetectID', 'istackImageID', 'ira', 'idec',
       'iraErr', 'idecErr', 'iEpoch', 'iPSFMag', 'iPSFMagErr', 'iApMag',
       'iApMagErr', 'iKronMag', 'iKronMagErr', 'iinfoFlag', 'iinfoFlag2',
       'iinfoFlag3', 'inFrames', 'ixPos', 'iyPos', 'ixPosErr', 'iyPosErr',
       'ipsfMajorFWHM', 'ipsfMinorFWHM', 'ipsfTheta', 'ipsfCore',
       'ipsfLikelihood', 'ipsfQf', 'ipsfQfPerfect', 'ipsfChiSq',
       'imomentXX', 'imomentXY', 'imomentYY', 'imomentR1', 'imomentRH',
       'iPSFFlux', 'iPSFFluxErr', 'iApFlux', 'iApFluxErr', 'iApFillFac',
       'iApRadius', 'iKronFlux', 'iKronFluxErr', 'iKronRad', 'iexpTime',
       'iExtNSigma', 'isky', 'iskyErr', 'izp', 'iPlateScale',
       'zippDetectID', 'zstackDetectID', 'zstackImageID', 'zra', 'zdec',
       'zraErr', 'zdecErr', 'zEpoch', 'zPSFMag', 'zPSFMagErr', 'zApMag',
       'zApMagErr', 'zKronMag', 'zKronMagErr', 'zinfoFlag', 'zinfoFlag2',
       'zinfoFlag3', 'znFrames', 'zxPos', 'zyPos', 'zxPosErr', 'zyPosErr',
       'zpsfMajorFWHM', 'zpsfMinorFWHM', 'zpsfTheta', 'zpsfCore',
       'zpsfLikelihood', 'zpsfQf', 'zpsfQfPerfect', 'zpsfChiSq',
       'zmomentXX', 'zmomentXY', 'zmomentYY', 'zmomentR1', 'zmomentRH',
       'zPSFFlux', 'zPSFFluxErr', 'zApFlux', 'zApFluxErr', 'zApFillFac',
       'zApRadius', 'zKronFlux', 'zKronFluxErr', 'zKronRad', 'zexpTime',
       'zExtNSigma', 'zsky', 'zskyErr', 'zzp', 'zPlateScale',
       'yippDetectID', 'ystackDetectID', 'ystackImageID', 'yra', 'ydec',
       'yraErr', 'ydecErr', 'yEpoch', 'yPSFMag', 'yPSFMagErr', 'yApMag',
       'yApMagErr', 'yKronMag', 'yKronMagErr', 'yinfoFlag', 'yinfoFlag2',
       'yinfoFlag3', 'ynFrames', 'yxPos', 'yyPos', 'yxPosErr', 'yyPosErr',
       'ypsfMajorFWHM', 'ypsfMinorFWHM', 'ypsfTheta', 'ypsfCore',
       'ypsfLikelihood', 'ypsfQf', 'ypsfQfPerfect', 'ypsfChiSq',
       'ymomentXX', 'ymomentXY', 'ymomentYY', 'ymomentR1', 'ymomentRH',
       'yPSFFlux', 'yPSFFluxErr', 'yApFlux', 'yApFluxErr', 'yApFillFac',
       'yApRadius', 'yKronFlux', 'yKronFluxErr', 'yKronRad', 'yexpTime',
       'yExtNSigma', 'ysky', 'yskyErr', 'yzp', 'yPlateScale', 'distance',
       'NED_name', 'NED_type', 'NED_vel', 'NED_redshift', 'NED_mag',
       'i-z', 'g-r', 'r-i', 'g-i', 'z-y', 'g-rErr', 'r-iErr', 'i-zErr',
       'z-yErr', 'gApMag_gKronMag', 'rApMag_rKronMag', 'iApMag_iKronMag',
       'zApMag_zKronMag', 'yApMag_yKronMag', '7DCD', 'class', 'dist/DLR',
       'dist', 'TransientClass', 'TransientRA', 'TransientDEC',
       'TransientDiscoveryDate', 'TransientDiscoveryMag',
       'TransientRedshift', 'TransientDiscoveryYear', 'Transient AltName',
       'host_logmass', 'host_logmass_min', 'host_logmass_max',
       'Hubble Residual', 'TransientName']
        df = pd.DataFrame(columns = colnames)
        df.to_csv("./database/GHOST.csv",index=False)
        if verbose:
            print("Successfully created dummy database.\n")

# Calculates the fraction of supernovae
# in our dictionary that have at least
# one candidate host associated with
# them.
# input: transient_dict, the dictionary
#        of supernovae and their host
#        galaxy candidate objIDs from PS1
# output: the fraction of supernovae
#         with candidate hosts
def fracWithHosts(transient_dict):
    count = 0
    for name, host in transient_dict.items():
        # only do matching if there's a found host
        if isinstance(host, list) or isinstance(host, np.ndarray):
            if len(host) > 0 and np.any(np.array(host == host)):
                count += 1
        #elif isinstance(host, np.ndarray):
        #    if len(host) > 0 and np.any(np.array(host == host)):
        #        count += 1
        else:
            if host == host:
                count+=1
    return count/len(transient_dict.keys())

# Removes the prefix from a string
# input - text and prefix to remove from text
# output - text without prefix
# very useful for removing the 'SN' from
# supernova names!
def remove_prefix(text, prefix):
    return text[text.startswith(prefix) and len(prefix):]

# Gets the host associated with a supernova in
# the GHOST database, if it exists. If not,
# this method runs through the pipeline to
# find a host for a new event, and returns all
# data for the transient host.
# input - TransientCoord, the coordinates of
#         a transient being queried
# output - A data frame of the host
#          associated with the transient
#          being queried, with PS1 information.
#          If the supernova is currently in the
#          database, return the database object.
#          If not, complete the pipeline for
#          host association and return the database
#          object.
#          If no host is found, return None

# Returns all data for a transient host
# currently in the database based on its
# coordinates.
# input - TransientCoord, the coordinates of
#         a new transient being queried
# output - A data frame of the host
#          associated with the transient
#          being queried, with PS1 information.
#          If the supernova is currently in the
#          database, return the database object.
#          If not found, return None
def getDBHostFromTransientCoords(transientCoords):
    fullTable = fullData()
    notFound = []
    host_DF = None
    hostList = []

    #a little wrapper so that this function can take lists of coords in addition to one coord
    for transientCoord in transientCoords:
        smallTable = fullTable[np.abs(fullTable['TransientRA'] - transientCoord.ra.degree)<0.1]
        smallTable = smallTable[np.abs(smallTable['TransientDEC'] - transientCoord.dec.degree)<0.1]
        if len(smallTable) < 1:
            notFound.append(transientCoord)
            continue
        c2 = SkyCoord(smallTable['TransientRA']*u.deg, smallTable['TransientDEC']*u.deg, frame='icrs')
        sep = np.array(transientCoord.separation(c2).arcsec)
        if np.nanmin(sep) <= 1:
            host_idx = np.where(sep == np.nanmin(sep))[0][0]
            host = smallTable.iloc[[host_idx]]
            hostList.append(host)
        else:
            notFound.append(transientCoord)
        #    print("Sorry, that supernova was not found in our database! The closest supernova is %.2f arcsec away.\n" %np.nanmin(sep))
    if len(hostList) > 0:
        host_DF = pd.concat(hostList, ignore_index=True)
    return host_DF, notFound

# Returns all data for transient host,
# based on supernova name
# input - TransientCoord, the coordinates of
#         a new transient being queried
# output - A data frame of the host
#          associated with the transient
#          being queried, with PS1 information.
#          If the supernova is currently in the
#          database, return the database object.
#          If no host is found, return None
def getDBHostFromTransientName(SNNames):
    fullTable = fullData()
    allHosts = []
    notFound = []
    host_DF = None
#    if isinstance(SNName, list) or isinstance(SNName, np.ndarray):
#        SNNames = SNName
#    else:
#        SNNames = np.array([SNname])
    for SNName in SNNames:
        SNName = re.sub(r"\s+", "", str(SNName))
        host = None
        possibleNames = [SNName, SNName.upper(), SNName.lower(), "SN"+SNName]
        for name in possibleNames:
            if len(fullTable[fullTable['TransientName'] == name])>0:
                host = fullTable[fullTable['TransientName'] == name]
                allHosts.append(host)
                break
        if host is None:
            notFound.append(SNName)
    if len(allHosts) > 0:
        host_DF = pd.concat(allHosts, ignore_index=True)
    return host_DF, notFound

# Returns all data for transient and host,
# based on host name
# input - hostName
# output - A data frame of the host
#          associated with the transient
#          being queried, with PS1 information.
#          If the supernova is currently in the
#          database, return the database object.
#          If no host is found, return None
def getHostFromHostName(hostNames):
    fullTable = fullData()
    hostName = np.array([re.sub(r"\s+", "", str(hostName)) for hostName in hostNames])
    possibleNames = []
    #allHosts = np.array([re.sub(r"\s+", "", str(x)) for x in fullTable['NED_name']])
    for name in hostName:
        possibleNames.append(name)
        possibleNames.append(name.upper())
        possibleNames.append(name.lower())

    host = fullTable[fullTable['NED_name'].isin(possibleNames)]
    if len(host) > 0:
        return host
    else:
        print("Sorry, no hosts were found in our database!\n")
        return None

def getHostFromHostCoords(hostCoords):
    fullTable = fullData()
    c2 = SkyCoord(fullTable['raMean']*u.deg, fullTable['decMean']*u.deg, frame='icrs')
    host = None
    for hostCoord in hostCoords:
        sep = np.array(hostCoords.separation(c2).arcsec)
        if np.nanmin(sep) <= 1:
            host_idx = np.where(sep == np.nanmin(sep))[0][0]
            host.append(fullTable.iloc[[host_idx]])
    if host is None:
        print("Sorry, No hosts found in our database!\n")
    else:
        host = pd.concat(host, ignore_index=True)
    return host

# Returns basic statistics for transient,
# based on a query of the coordinates of
# its host
# inputs - the position of the host
# outputs - A printout of name,
#           discovery year and discovery mag
#           for all transients associated
#           with a host at that coordinates

def getTransientStatsFromHostCoords(hostCoord):
    host = getHostFromHostCoords(hostCoord)
    i = 0
    if len(host) > 0:
        print("Found info for host %s.\n"%host['NED_name'].values[0])
        print("Associated supernovae: ")
        for SN in np.array(host['TransientName'].values):
            SN_frame = host.loc[host['TransientName'] == SN]
            print("%i. %s"%(i+1, SN_frame['TransientName'].values[0]))
            print("RA, DEC (J2000): %f, %f"%(SN_frame['TransientRA'].values[0], SN_frame['TransientDEC'].values[0]))
            print("Redshift: %f"%SN_frame['TransientRedshift'].values[0])
            print("Discovery Date: %s"%SN_frame['TransientDiscoveryDate'].values[0].split(" ")[0])
            print("Discovery Mag: %.2f"%SN_frame['TransientDiscoveryMag'].values[0])
            i+= 1
    return

# Returns basic statistics for transient,
# based on a query of the coordinates of
# its host
# inputs - the position of the host
# outputs - A printout of name,
#           discovery year and discovery mag
#           for all transients associated
#           with a host at that coordinates
def getTransientStatsFromHostName(hostName):
    host = getHostFromHostName(hostName)
    hostCoord = SkyCoord(np.unique(host['raMean'])*u.deg, np.unique(host['decMean'])*u.deg, frame='icrs')
    getTransientStatsFromHostCoords(hostCoord)
    return

# Returns basic statistics for the most likely
# host of a previously identified transient, using
# the pipeline outlined above.
# inputs - the coordinates of the transient to search
def getHostStatsFromTransientCoords(transientCoordsList):
    fullTable = fullData()
    names = []
    for transientCoords in transientCoordsList:
        c2 = SkyCoord(fullTable['TransientRA']*u.deg, fullTable['TransientDEC']*u.deg, frame='icrs')
        sep = np.array(transientCoords.separation(c2).arcsec)
        host = None
        if np.nanmin(sep) <= 1:
            host_idx = np.where(sep == np.nanmin(sep))[0][0]
            host = fullTable.iloc[[host_idx]]
            names.append(host['TransientName'].values[0])
    getHostStatsFromTransientName(names)

# Returns basic statistics for the most likely
# host of a previously identified transient, using
# the pipeline outlined above.
# inputs - the coordinates of the transient to search
def getHostStatsFromTransientName(SNName):
    SNName = np.array(SNName)
    fullTable = fullData()
    host, notFound = getDBHostFromTransientName(SNName)
    if host is not None:
        for idx, row in host.iterrows():
            if np.unique(row['NED_name']) != "":
                print("Found host %s.\n"%row['NED_name'])
            else:
                print("Found host with PS1 ID %s"%row['objID'])
            print("RA, DEC (J2000): %f, %f"%(row['raMean'], row['decMean']))
            if row['NED_redshift'] != '':
                print("Redshift: %f"%row['NED_redshift'])
            print("PS1 rMag: %.2f"%row['rApMag'])
            print("g-r          r-i          i-z")
            print("%.2f+/-%.3f   %.2f+/-%.3f   %.2f+/-%.3f"%(row['g-r'],row['g-rErr'], row['r-i'], row['r-iErr'], row['i-z'], row['i-zErr']))
        #    print("g-r: %.2f."%host['g-r'].values[0])
        #    print("r-i: %.2f."%host['r-i'].values[0])
        #    print("i-z: %.2f."%host['i-z'].values[0])
            #print("There are %i supernovae associated with this object.\n"%len(host))
            print("Associated supernovae: ")
            if np.unique(row['NED_name']) != "":
                tempHost = fullTable[fullTable['NED_name'] == row['NED_name']]
            else:
                tempHost = fullTable[fullTable['objID'] == row['objID']]
            for SN in np.array(tempHost['TransientName'].values):
                print(SN)
    else:
        print("No host info found!")
    return

# Returns a postage stamp of the most likely
# host in one of the PS1 bands - g,r,i,z,y - as a
# fits file with radius rad, and plots the image.
# inputs
# outputs
def getHostImage(transientName='', band="grizy", rad=60, save=0):
    if transientName == '':
        print("Error! Please enter a supernova!\n")
        return
    fullTable = fullData()
    host = getDBHostFromTransientName(transientName)
    host.reset_index(drop=True, inplace=True)
    tempSize = int(4*float(rad))
    fn_save = host['objID'].values[0]
    if np.unique(host['NED_name']) != "":
        fn_save = host['NED_name'].values[0]
        print("Showing postage stamp for %s"%np.unique(host['NED_name'])[0])
    ra = np.unique(host['raMean'])[0]
    dec = np.unique(host['decMean'])[0]
    tempSize = int(4*rad)
    img = getcolorim(ra, dec, output_size=tempSize, size=tempSize, filters=band, format="png")
    plt.figure(figsize=(10,10))
    plt.imshow(img)
    if save:
        img.save("%s.png" % fn_save)
    return

# description
# inputs
# outputs
def getTransientSpectra(path, SNname):#
    SNname = remove_prefix(SNname, 'SN')
    files = glob.glob(path+"*%s*"%SNname)
    specFiles = []
    if len(files) > 0:
        for file in files:
            if remove_prefix(file, path).startswith("osc_"):
                spectra = pd.read_csv(file, header=None, names=['Wave(A)', 'I', 'Ierr'])
            else:
                spectra = pd.read_csv(file, delim_whitespace=True, header=None, names=['x', 'y', 'z'])
            #spectra = spectra.astype('float')
            specFiles.append(spectra)
    else:
        print("Sorry! No spectra found.")
    return specFiles

# description
# inputs
# outputs
def getHostSpectra(SNname, path):#
    SNname = remove_prefix(SNname, 'SN')
    files = glob.glob(path+"*%s_hostSpectra.csv*"%SNname)
    specFiles = []
    if len(files) > 0:
        for file in files:
            spectra = pd.read_csv(file)
            #spectra = spectra.astype('float')
            specFiles.append(spectra)
    else:
        print("Sorry! No spectra found.")
    return specFiles

# A cone search for all transient-host pairs
# within a certain radius, returned as a pandas dataframe
# inputs location, in astropy coordinates, and radius in arcsec
# outputs the data frame of all nearby pairs
def coneSearchPairs(coord, radius):
    fullTable = fullData()
    c2 = SkyCoord(fullTable['TransientRA']*u.deg, fullTable['TransientDEC']*u.deg, frame='icrs')
    sep = np.array(coord.separation(c2).arcsec)
    hosts = None
    if np.nanmin(sep) < radius:
        host_idx = np.where(sep < radius)[0]
        hosts = fullTable.iloc[host_idx]
        #hosts = pd.concat(hosts)
    return hosts

# Returns the full table of data
# inputs: none
# outputs: the full GHOST database
def fullData():
    path = sys.path[0]
    fullTable = pd.read_csv(path+"/database/GHOST.csv")
    return fullTable

# The wrapper function for the
# host association pipeline.
# inputs: the location of the supernova
#         as an astropy SkyCoord object
# output: A pandas dataframe of
#         the most likely host in PS1,
#         with stats provided at
#         printout
def getTransientHosts(snName=[''], snCoord=[''], snClass=[''], verbose=0, starcut='normal'):
    hostDB = None
    if not isinstance(snCoord, list) and not isinstance(snCoord, np.ndarray):
        snCoord = [snCoord]
        snName = [snName]
        snClass = [snClass]
    if len(snClass) != len(snName):
        snClass = ['']*len(snName)

    snName = [x.replace(" ", "") for x in snName]
    df_transients = pd.DataFrame({'Name':np.array(snName), 'snCoord':np.array(snCoord), 'snClass':np.array(snClass)})

    tempHost1, notFoundNames = getDBHostFromTransientName(snName)
    found_by_name = len(snName) - len(notFoundNames)
    if tempHost1 is None or len(notFoundNames) > 0:
        if verbose:
            print("%i transients not found in GHOST by name, trying a coordinate search..."%len(notFoundNames))

        df_transients_remaining = df_transients[df_transients['Name'].isin(notFoundNames)]

        snCoord_remaining =  df_transients_remaining['snCoord'].values

        tempHost2, notFoundCoords = getDBHostFromTransientCoords(snCoord_remaining);
        found_by_coord = len(snName) - len(notFoundCoords) - found_by_name
        if tempHost2 is None or len(notFoundCoords) > 0:
            if verbose:
                print("%i transients not found in GHOST by name or coordinates, manually associating..."% len(notFoundCoords))

            df_transients_remaining = df_transients_remaining[df_transients_remaining['snCoord'].isin(notFoundCoords)]

            snName_remaining =  df_transients_remaining['Name'].values
            snCoord_remaining = df_transients_remaining['snCoord'].values
            snClass_remaining = df_transients_remaining['snClass'].values

            tempHost3 = findNewHosts(snName_remaining, snCoord_remaining, snClass_remaining, verbose, starcut)
            found_by_manual = len(tempHost3)
    hostDB = pd.concat([tempHost1, tempHost2, tempHost3], ignore_index=True)
    hostDB.replace(-999.0, np.nan, inplace=True)
    if verbose:
        print("%i transients found by name, %i transients found by coordinates, %i transients manually associated."% (found_by_name, found_by_coord, found_by_manual))
    return hostDB

# The wrapper function for the
# host association pipeline.
# inputs: the location of the supernova
#         as an astropy SkyCoord object
# output: A pandas dataframe of
#         the most likely host in PS1,
#         with stats provided at
#         printout
def findNewHosts(snName, snCoord, snClass, verbose=0, starcut='gentle'):
    if isinstance(snName, str):
        snName = snName.replace(" ", "")
        snRA = snCoord.ra.degree
        snDEC = snCoord.dec.degree
    elif isinstance(snName, list) or isinstance(snName, np.ndarray):
        snName = [x.replace(" ", "") for x in snName]
        snRA = [x.ra.degree for x in snCoord]
        snDEC = [x.dec.degree for x in snCoord]
    else:
        print("Error! Please pass in your transient name as either a string or a list/array of strings.\n")
        #return None

    snName_arr = np.array(snName)
    snRA_arr = np.array(snRA)
    snDEC_arr = np.array(snDEC)
    snClass_arr = np.array(snClass)

    #now = datetime.now()
    #dateStr = "%i%.02i%.02i" % (now.year,now.month,now.day)
    dateStr = str(datetime.today()).replace("-", '').replace(".", '').replace(":", "").replace(" ", '')

    rad = 30 #arcsec
    fn_Host = "SNe_TNS_%s_PS1Hosts_%iarcsec.csv" % (dateStr, rad)
    fn_SN = 'transients_%s.csv' % dateStr
    fn_Dict = fn_Host[:-4] + ".p"
    dir_name = fn_SN[:-4]
    if not os.path.exists('./' + dir_name):
        os.makedirs('./' + dir_name)
    path = './'+dir_name+'/'
    #create temp dataframe with RA and DEC corresponding to the transient

    snDF = pd.DataFrame({'Name':snName_arr, 'RA':snRA_arr, 'DEC':snDEC_arr, 'HostName':['']*len(snDEC_arr), 'Obj. Type':snClass_arr})
    snDF.to_csv(path+fn_SN, index=False)

    #begin doing the heavy lifting to associate transients with hosts
    host_DF = get_hosts(path, fn_SN, fn_Host, rad)

    if len(host_DF) < 1:
        print("ERROR: Found no hosts in cone search during manual association!")
        return None

    cuts = ["n", "quality", "coords", "primary", "best", "duplicate"]

    transient_dict = pickle.load(open(path+'/dictionaries/'+fn_Dict, "rb"))
    transient_dict = {k.replace(' ', ''): v for k, v in transient_dict.items()}
    # check how many supernovae we have potential hosts for - should be nearly all of em

    host_DF = makeCuts(host_DF, cuts, transient_dict)
    cut_dict = clean_dict(transient_dict, host_DF, [])
    hostFrac = fracWithHosts(cut_dict)*100
    if verbose:
        print("Associated fraction after quality cuts: %.2f%%."%hostFrac)

    #automatically add to host association for gradient ascent
    lost = np.array([k for k, v in cut_dict.items() if len(v) <1])
    #lost = set(snName_arr) - set(cut_dict.keys())

    host_DF = getColors(host_DF)
    host_DF = removePS1Duplicates(host_DF)

    host_DF = getNEDInfo(host_DF)
    host_DF = calc_7DCD(host_DF)
    host_gals_DF, stars_DF = separateStars_STRM(host_DF, plot=0, verbose=verbose, starcut=starcut)
    if verbose:
        print("Removed %i stars. We now have %i candidate host galaxies."%(len(stars_DF), len(host_gals_DF)))

    cut_dict = clean_dict(cut_dict, host_gals_DF, [])
    stars_DF.to_csv(path+"removedStars.csv",index=False)

    #plotLocus(host_gals_DF, color=1, save=1, type="Gals", timestamp=dateStr)
    #plotLocus(stars_DF, color=1, save=1, type="Stars", timestamp=dateStr)

    host_dict_nospace = {k.replace(' ', ''): v for k, v in cut_dict.items()}

    fn = "DLR.txt"
    transients = pd.read_csv(path+fn_SN)
#    clean_dict(host_dict_nospace, host_gals_DF, [])
#    fracWithHosts(host_dict_nospace)

    with open(path+"/dictionaries/checkpoint_preDLR.p", 'wb') as fp:
           pickle.dump(host_dict_nospace, fp, protocol=pickle.HIGHEST_PROTOCOL)

    host_DF, host_dict_nospace_postDLR, noHosts, GD_SN = chooseByDLR(path, host_gals_DF, transients, fn, host_dict_nospace, host_dict_nospace, todo="r")

    if len(noHosts) > 0:
        with open(path+"/noHosts_fromDLR_griCut.txt", 'wb') as fp:
            pickle.dump(noHosts, fp)

    if len(GD_SN) > 0:
        with open(path+"/badInfo_fromDLR_griCut.txt", 'wb') as fp:
            pickle.dump(GD_SN, fp)

    #gradient ascent algorithm for the SNe that didn't pass this stage
    SN_toReassociate = np.concatenate([np.array(noHosts), np.array(GD_SN), np.array(list(lost))])
    SN_toReassociate = np.array(SN_toReassociate)

    if len(SN_toReassociate) > 0:
        if verbose:
            print("%i transients with no host found with DLR, %i transients with bad host data with DLR." %(len(noHosts), len(GD_SN)))
            print("Running gradient ascent for %i remaining transients..."%len(SN_toReassociate))
            print("See GradientAscent.txt for more information.")

        fn_GD= path+'/GradientAscent.txt'

        host_dict_nospace_postDLR_GD, host_DF, unchanged = gradientAscent(path, transient_dict,  host_dict_nospace_postDLR, SN_toReassociate, host_DF, transients, fn_GD, plot=verbose)

        with open(path+"/dictionaries/gals_postGD.p", 'wb') as fp:
            pickle.dump(host_dict_nospace_postDLR_GD, fp, protocol=pickle.HIGHEST_PROTOCOL)

        if verbose:
            print("Hosts not found for %i transients in gradient ascent. Storing names in GD_unchanged.txt" %(len(unchanged)))

        with open(path+"/GD_unchanged.txt", 'wb') as fp:
            pickle.dump(unchanged, fp)

        hostFrac = fracWithHosts(host_dict_nospace_postDLR_GD)*100

        if verbose:
            print("Associated fraction after gradient ascent: %.2f%%."%hostFrac)

        final_dict = host_dict_nospace_postDLR_GD.copy()

    else:
        final_dict = host_dict_nospace_postDLR.copy()

    host_DF = build_ML_df(final_dict, host_DF, transients)

    with open(path+"/dictionaries/" + "Final_Dictionary.p", 'wb') as fp:
           pickle.dump(final_dict, fp, protocol=pickle.HIGHEST_PROTOCOL)

    #a few final cleaning steps
    host_DF.drop_duplicates(subset=['TransientName'], inplace=True)
    host_DF = host_DF[host_DF['TransientName'] != ""]
    host_DF.reset_index(inplace=True, drop=True)
    host_DF['TransientName'] = [x.replace(" ", "") for x in host_DF['TransientName']]

    matchFrac = len(host_DF)/len(transients)*100
    print("Found matches for %.1f%% of events."%matchFrac)
    if verbose:
        print("Saving table of hosts to %s."%(path+"tables/FinalAssociationTable.csv"))

    host_DF.to_csv(path+"FinalAssociationTable.csv", index=False)

    #hSpecPath = path+"/hostSpectra/"
    #tSpecPath = path+"/SNspectra/"
    #psPath = path+"/hostPostageStamps/"
    tablePath = path+"/tables/"
    printoutPath = path+'/printouts/'

    #paths = [hSpecPath, tSpecPath, psPath, tablePath, printoutPath]
    paths = [tablePath, printoutPath]

    for tempPath in paths:
        if not os.path.exists(tempPath):
            os.makedirs(tempPath)

    #move tables to the tables directory, and printouts to the printouts directory
    printouts = glob.glob(path+'/*.txt')
    for p in printouts:
        fn = remove_prefix(p, path)
        os.rename(p, printoutPath+fn)

    table_ext = ['csv', 'gz']
    for end in table_ext:
        tables = glob.glob(path+'/*.%s'%end)
        for t in tables:
            fn = remove_prefix(t, path)
            os.rename(t, tablePath+fn)

    return(host_DF)
